# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import typing as tp
import os
import numpy as np
import pyomo.environ as pyomo
import nevergrad as ng
from . import core


def test_concrete_model_without_constraints() -> None:
    model = pyomo.ConcreteModel()
    model.x = pyomo.Var([1, 2], domain=pyomo.NonNegativeReals)
    model.obj = pyomo.Objective(expr=(model.x[1] - 0.5) ** 2 + (model.x[2] - 0.5) ** 2)

    func = core.Pyomo(model)
    optimizer = ng.optimizers.NGO(parametrization=func.parametrization, budget=100)
    recommendation = optimizer.minimize(func.function)

    np.testing.assert_almost_equal(recommendation.kwargs["x[1]"], 0.5, decimal=1)
    np.testing.assert_almost_equal(recommendation.kwargs["x[2]"], 0.5, decimal=1)


def square(m: tp.Any) -> float:
    return pyomo.quicksum((m.x[i] - 0.5) ** 2 for i in m.x)


def test_concrete_model_with_constraints() -> None:
    model = pyomo.ConcreteModel()
    model.x = pyomo.Var([0, 1], domain=pyomo.Reals)
    model.obj = pyomo.Objective(rule=square)
    model.Constraint1 = pyomo.Constraint(rule=lambda m: m.x[0] >= 1)
    model.Constraint2 = pyomo.Constraint(rule=lambda m: m.x[1] >= 0.8)

    func = core.Pyomo(model)
    optimizer = ng.optimizers.OnePlusOne(parametrization=func.parametrization, budget=100)
    recommendation = optimizer.minimize(func.function)

    np.testing.assert_almost_equal(recommendation.kwargs["x[0]"], 1.0, decimal=1)
    np.testing.assert_almost_equal(recommendation.kwargs["x[1]"], 0.8, decimal=1)


def test_abstract_model_with_constraints() -> None:
    abs_model = pyomo.AbstractModel()
    abs_model.F = pyomo.Set()
    abs_model.Xmin = pyomo.Param(abs_model.F, within=pyomo.Reals, default=0.0)
    abs_model.x = pyomo.Var(abs_model.F, within=pyomo.Reals)
    abs_model.constraints = pyomo.Constraint(abs_model.F, rule=lambda m, i: m.x[i] >= m.Xmin[i])
    abs_model.obj = pyomo.Objective(rule=square)

    # Load the values of the parameters from external file
    dirname = os.path.dirname(__file__)
    filename = os.path.join(dirname, "test_model_1.dat")
    model = abs_model.create_instance(filename)

    func = core.Pyomo(model)
    func.parametrization.random_state.seed(12)
    optimizer = ng.optimizers.OnePlusOne(parametrization=func.parametrization, budget=200)
    recommendation = optimizer.minimize(func.function)

    np.testing.assert_almost_equal(recommendation.kwargs['x["New York"]'], model.Xmin["New York"], decimal=1)
    np.testing.assert_almost_equal(
        recommendation.kwargs['x["Hong Kong"]'], model.Xmin["Hong Kong"], decimal=1
    )


def test_pyomo_set() -> None:
    def square2(m: tp.Any) -> float:
        return (m.x - 1) ** 2  # type: ignore

    model = pyomo.ConcreteModel()
    model.P = pyomo.Set(initialize=list(range(1, 11)))
    model.Q = pyomo.Set(initialize=list(range(6, 16)))
    model.R = model.P ^ model.Q  # XOR
    model.x = pyomo.Var(domain=model.R)
    model.obj = pyomo.Objective(rule=square2)
    model.constraint1 = pyomo.Constraint(rule=lambda m: m.x >= 2)

    func = core.Pyomo(model)
    func.parametrization.random_state.seed(12)
    optimizer = ng.optimizers.OnePlusOne(parametrization=func.parametrization, budget=100)
    recommendation = optimizer.minimize(func.function)

    np.testing.assert_almost_equal(recommendation.kwargs["x"], 2.0, decimal=1)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import typing as tp
from nevergrad.functions import gym as nevergrad_gym
from .xpbase import registry
from .xpbase import create_seed_generator
from .xpbase import Experiment
from .optgroups import get_optimizers

# pylint: disable=too-many-nested-blocks,stop-iteration-return


# A few GYM modifiers based on environment variables.
def gym_problem_modifier(specific_problem):
    specific_problem = os.environ.get("TARGET_GYM_ENV", specific_problem)
    print("problem=", specific_problem)
    return specific_problem


def gym_optimizer_modifier(optims):
    print(optims)
    if os.environ.get("GYM_OPTIMIZER") is not None:
        optimizer_string = os.environ.get("GYM_OPTIMIZER")
        print(f"Considering optimizers with {optimizer_string} in their name.")
        optims = [o for o in optims if optimizer_string in str(o)]  # type: ignore
        if len(optims) == 0:
            optims = [optimizer_string]  # type: ignore
    print("optims=", optims)
    return optims


def gym_budget_modifier(budgets):
    if os.environ.get("MAX_GYM_BUDGET") is not None:
        budget_string = os.environ.get("MAX_GYM_BUDGET")
        budgets = [b for b in budgets if b < int(budget_string)]  # type: ignore
    print("budgets=", budgets)
    return budgets


@registry.register
def ng_full_gym(
    seed: tp.Optional[int] = None,
    randomized: bool = True,
    multi: bool = False,
    big: bool = False,
    memory: bool = False,
    ng_gym: bool = False,  # pylint: disable=redefined-outer-name
    conformant: bool = False,
    gp: bool = False,
    sparse: bool = False,
    multi_scale: bool = False,
) -> tp.Iterator[Experiment]:
    """Gym simulator. Maximize reward.  Many distinct problems.

    Parameters:
        seed: int
           random seed.
        randomized: bool
           whether we keep the problem's stochasticity
        multi: bool
           do we have one neural net per time step
        big: bool
           do we consider big budgets
        memory: bool
           do we use recurrent nets
        ng_gym: bool
           do we restrict to ng-gym
        conformant: bool
           do we restrict to conformant planning, i.e. deterministic controls.
    """
    env_names = nevergrad_gym.GymMulti.get_env_names()
    assert int(ng_gym) + int(gp) <= 1, "At most one specific list of environments."
    if ng_gym:
        env_names = nevergrad_gym.GymMulti.ng_gym
    if gp:
        import pybullet_envs  # pylint: disable=unused-import

        env_names = [
            "CartPole-v1",
            "Acrobot-v1",
            "MountainCarContinuous-v0",
            "Pendulum-v1",
            # "InvertedPendulumSwingupBulletEnv-v0",
            "BipedalWalker-v3",
            "BipedalWalkerHardcore-v3",
            "HopperBulletEnv-v0",
            # "InvertedDoublePendulumBulletEnv-v0",
            "LunarLanderContinuous-v2",
        ]

    seedg = create_seed_generator(seed)
    optims = [
        "DiagonalCMA",
        "GeneticDE",
        "NoisyRL1",
        "NoisyRL2",
        "NoisyRL3",
        "MixDeterministicRL",
        "SpecialRL",
        "PSO",
    ]
    if multi:
        controls = ["multi_neural"]
    else:
        controls = (
            [
                "noisy_semideep_neural",
                "noisy_scrambled_semideep_neural",  # Scrambling: why not perturbating the order of variables ?
                "noisy_deep_neural",
                "noisy_scrambled_deep_neural",
                "neural",
                # "structured_neural",
                # "memory_neural",
                "stackingmemory_neural",
                "deep_neural",
                "semideep_neural",
                "noisy_neural",
                "noisy_scrambled_neural",
                # "scrambled_neural",
                # "linear",
                "resid_neural",
                "resid_semideep_neural",
                "resid_deep_neural",
            ]
            if not big
            else ["resid_neural"]
        )
    if memory:
        controls = ["stackingmemory_neural", "deep_stackingmemory_neural", "semideep_stackingmemory_neural"]
        controls += ["memory_neural", "deep_memory_neural", "semideep_memory_neural"]
        controls += [
            "extrapolatestackingmemory_neural",
            "deep_extrapolatestackingmemory_neural",
            "semideep_extrapolatestackingmemory_neural",
        ]
        assert not multi
    if conformant:
        controls = ["stochastic_conformant"]
    optimization_scales: tp.List[int] = [0]
    if multi_scale:
        optimization_scales = [-6, -4, -2, 0]
    budgets = [50, 200, 100, 25, 400]  # Let's go with low budget.
    budgets = gym_budget_modifier(budgets)
    for control in controls:
        neural_factors: tp.Any = (
            [None]
            if (conformant or control == "linear")
            else ([1] if "memory" in control else ([3] if big else [1, 2, 3]))
        )
        for neural_factor in neural_factors:
            for name in env_names:
                sparse_limits: tp.List[tp.Optional[int]] = [None]
                if sparse:
                    sparse_limits += [10, 100, 1000]
                for sparse_limit in sparse_limits:
                    for optimization_scale in optimization_scales:
                        try:
                            func = nevergrad_gym.GymMulti(
                                name,
                                control=control,
                                neural_factor=neural_factor,
                                randomized=randomized,
                                optimization_scale=optimization_scale,
                                sparse_limit=sparse_limit,
                            )
                        except MemoryError:
                            continue
                        for budget in budgets:
                            for algo in optims:
                                xp = Experiment(func, algo, budget, num_workers=1, seed=next(seedg))
                                if not xp.is_incoherent:
                                    yield xp


@registry.register
def multi_ng_full_gym(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    """Counterpart of ng_full_gym with one neural net per time step.

    Each neural net is used for many problems, but only for one of the time steps."""
    return ng_full_gym(seed, multi=True)


@registry.register
def conformant_ng_full_gym(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    """Counterpart of ng_full_gym with fixed, predetermined actions for each time step.

    This is conformant: we optimize directly the actions for a given context.
    This does not prevent stochasticity, but actions do not depend on observationos.
    """
    return ng_full_gym(seed, conformant=True)


@registry.register
def ng_gym(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    """Counterpart of ng_full_gym with a specific, reduced list of problems."""
    return ng_full_gym(seed, ng_gym=True)


@registry.register
def gp(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    """GP benchmark.

    Counterpart of ng_full_gym with a specific, reduced list of problems for matching
    a genetic programming benchmark."""
    return ng_full_gym(seed, gp=True, multi_scale=True)


@registry.register
def conformant_gp(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    """GP benchmark.

    Counterpart of ng_full_gym with a specific, reduced list of problems for matching
    a genetic programming benchmark."""
    return ng_full_gym(seed, conformant=True, gp=True)


@registry.register
def sparse_gp(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    """GP benchmark.

    Counterpart of ng_full_gym with a specific, reduced list of problems for matching
    a genetic programming benchmark."""
    return ng_full_gym(seed, gp=True, sparse=True, multi_scale=True)


@registry.register
def ng_stacking_gym(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    """Counterpart of ng_gym with a recurrent network."""
    return ng_full_gym(seed, ng_gym=True, memory=True)


@registry.register
def big_gym_multi(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    """Counterpart of ng_full_gym with bigger nets."""
    return ng_full_gym(seed, big=True)


@registry.register
def deterministic_gym_multi(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    """Counterpart of ng_full_gym with fixed seeds (so that the problem becomes deterministic)."""
    return ng_full_gym(seed, randomized=False)


# Not registered because not validated.
def gym_multifid_anm(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    """Gym simulator for Active Network Management."""

    func = nevergrad_gym.GymMulti("multifidLANM")
    seedg = create_seed_generator(seed)
    optims = get_optimizers("basics", "progressive", "splitters", "baselines", seed=next(seedg))
    for budget in [25, 50, 100, 200, 400, 800, 1600, 3200, 6400, 12800, 25600, 51200]:
        for num_workers in [1, 30]:
            if num_workers < budget:
                for algo in optims:
                    xp = Experiment(func, algo, budget, num_workers=num_workers, seed=next(seedg))
                    if not xp.is_incoherent:
                        yield xp


def gym_problem(
    seed: tp.Optional[int] = None,
    specific_problem: str = "LANM",
    conformant: bool = False,
    big_noise: bool = False,
    multi_scale: bool = False,
    greedy_bias: bool = False,
) -> tp.Iterator[Experiment]:
    """Gym simulator for Active Network Management (default) or other pb.

    seed: int
        random seed for determinizing the problem
    specific_problem: string
        name of the problem we are working on
    conformant: bool
        do we focus on conformant planning
    big_noise: bool
        do we switch to specific optimizers, dedicated to noise
    multi_scale: boolean
        do we check multiple scales
    greedy_bias: boolean
        do we use greedy reward estimates for biasing the decisions.
    """
    if conformant:
        funcs = [
            nevergrad_gym.GymMulti(
                specific_problem,
                control="conformant",
                neural_factor=None,
            )
        ]
    else:
        funcs = [
            nevergrad_gym.GymMulti(
                specific_problem,
                control=control,
                neural_factor=1 if control != "linear" else None,
                optimization_scale=scale,
                greedy_bias=greedy_bias,
            )
            for scale in ([-6, -4, -2, 0] if multi_scale else [0])
            for control in (
                ["deep_neural", "semideep_neural", "neural", "linear"] if not greedy_bias else ["neural"]
            )
        ]
    seedg = create_seed_generator(seed)
    optims = [
        "TwoPointsDE",
        "GeneticDE",
        "PSO",
        "DiagonalCMA",
        "DoubleFastGADiscreteOnePlusOne",
        "DiscreteLenglerOnePlusOne",
        "PortfolioDiscreteOnePlusOne",
        "MixDeterministicRL",
        "NoisyRL2",
        "NoisyRL3",
        "SpecialRL",
        "NGOpt39",
        "CMA",
        "DE",
    ]
    if "stochastic" in specific_problem:
        optims = ["DiagonalCMA", "TBPSA"] if big_noise else ["DiagonalCMA"]
    if specific_problem == "EnergySavingsGym-v0" and conformant:  # Do this for all conformant discrete ?
        optims = [
            "DiscreteOnePlusOne",
            "PortfolioDiscreteOnePlusOne",
            "DiscreteLenglerOnePlusOne",
            "AdaptiveDiscreteOnePlusOne",
            "AnisotropicAdaptiveDiscreteOnePlusOne",
            "DiscreteBSOOnePlusOne",
            "DiscreteDoerrOnePlusOne",
            "OptimisticDiscreteOnePlusOne",
            "NoisyDiscreteOnePlusOne",
            "DoubleFastGADiscreteOnePlusOne",
            "SparseDoubleFastGADiscreteOnePlusOne",
            "RecombiningPortfolioOptimisticNoisyDiscreteOnePlusOne",
            "RecombiningPortfolioDiscreteOnePlusOne",
            "MultiDiscrete",
            "NGOpt",
        ]

    optims = gym_optimizer_modifier(optims)
    budgets = [25, 50, 100, 200]
    budgets = gym_budget_modifier(budgets)
    for func in funcs:
        for budget in budgets:
            for num_workers in [1]:
                if num_workers < budget:
                    for algo in optims:
                        xp = Experiment(func, algo, budget, num_workers=num_workers, seed=next(seedg))
                        if not xp.is_incoherent:
                            yield xp


@registry.register
def conformant_planning(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    # You might modify this problem by specifying an environment variable.
    specific_problem = "EnergySavingsGym-v0"

    return gym_problem(
        seed,
        specific_problem=gym_problem_modifier(specific_problem),
        conformant=True,
        big_noise=False,
    )


@registry.register
def neuro_planning(seed: tp.Optional[int] = None) -> tp.Iterator[Experiment]:
    # You might modify this problem by specifying an environment variable.
    specific_problem = "EnergySavingsGym-v0"
    return gym_problem(
        seed,
        specific_problem=gym_problem_modifier(specific_problem),
        conformant=False,
        big_noise=False,
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import warnings
import contextlib
import itertools
import numpy as np
import nevergrad.common.typing as tp
from . import core
from . import container
from . import _layering
from . import _datalayers
from . import data as pdata
from . import choice as pchoice
from . import transforms as trans


def flatten(
    parameter: core.Parameter, with_containers: bool = True, order: int = 0
) -> tp.List[tp.Tuple[str, core.Parameter]]:
    """List all the instances involved as parameter (not as subparameter/
    endogeneous parameter)

    Parameter
    ---------
    parameter: Parameter
        the parameter to inspect
    with_containers: bool
        returns all the sub-parameter if True, otherwise only non-pure containers (i.e. only Data and Choice)
    order: int
        order of model/internal parameters to extract. With 0, no model/internal parameters is
        extracted, with 1, only 1st order are extracted, with 2, so model/internal parameters and
        their own model/internal parameters etc. Order 1 subparameters/endogeneous parameters
        include sigma for instance.


    Returns
    -------
    list
        a list of all (name, parameter) implied in this parameter, i.e all choices, items of dict
        and tuples etc (except if only_data=True). Names have a format "<index>.<key>" for a tuple
        containing dicts containing data for instance. Supbaramaters have # in their names.
        The parameters are sorted in the same way they would appear in the standard data.

    """
    flat = [("", parameter)]
    if isinstance(parameter, container.Container):
        content_to_add: tp.List[container.Container] = [parameter]
        if isinstance(parameter, container.Instrumentation):  # special case: skip internal Tuple and Dict
            content_to_add = [parameter[0], parameter[1]]  # type: ignore
        for c in content_to_add:
            for k, p in sorted(c._content.items()):
                content = flatten(p, with_containers=with_containers, order=order)
                flat.extend(
                    (str(k) + ("" if not x else ("." if not x.startswith("#") else "") + x), y)
                    for x, y in content
                )
    if order > 0 and not isinstance(parameter, container.Container):
        subcontent = dict(parameter._subobjects.items())
        param = container.Dict(**subcontent)
        if len(subcontent) == 1:
            lone_content = next(iter(subcontent.values()))
            if isinstance(lone_content, container.Dict):
                param = lone_content  # shorcut subparameters
        subparams = flatten(param, with_containers=False, order=order - 1)
        flat += [(f"#{x}", y) for x, y in subparams]
    if not with_containers:
        flat = [(x, y) for x, y in flat if isinstance(y, (pdata.Data, pchoice.BaseChoice))]
    return flat


def list_data(parameter: core.Parameter) -> tp.List[pdata.Data]:
    """List all the Data instances involved as parameter (not as subparameter/
    endogeneous parameter) in the order they are defined in the standardized data.

    Parameter
    ---------
    parameter: Parameter
        the parameter to inspect
    """
    return [x for _, x in flatten(parameter, order=0) if isinstance(x, pdata.Data)]


class ParameterInfo(tp.NamedTuple):
    """Information about a parameter

    Attributes
    ----------
    deterministic: bool
        whether the function equipped with its instrumentation is deterministic.
        Can be false if the function is not deterministic or if the instrumentation
        contains a softmax.
    continuous: bool
        whether the domain is entirely continuous.
    ordered: bool
        whether all domains and subdomains are ordered.
    arity: int
        number of options for discrete parameters (-1 if continuous)
    """

    deterministic: bool
    continuous: bool
    ordered: bool
    arity: int


def analyze(parameter: core.Parameter) -> ParameterInfo:
    """Analyzes a parameter to provide useful information about it"""
    params = list_data(parameter)
    int_layers = list(itertools.chain.from_iterable([_layering.Int.filter_from(x) for x in params]))
    return ParameterInfo(
        deterministic=all(lay.deterministic for lay in int_layers),
        continuous=not any(lay.deterministic for lay in int_layers),
        ordered=all(lay.ordered for lay in int_layers),
        arity=max(
            (lay.arity for lay in int_layers if lay.arity is not None), default=-1
        ),  # only softmax params for now
    )


@contextlib.contextmanager
def deterministic_sampling(parameter: core.Parameter) -> tp.Iterator[None]:
    """Temporarily change the behavior of a Parameter to become deterministic

    Parameters
    ----------
    parameter: Parameter
        the parameter which must behave deterministically during the "with" context
    """
    all_data = list_data(parameter)
    int_layers = list(itertools.chain.from_iterable([_layering.Int.filter_from(x) for x in all_data]))
    # record state and set deterministic to True
    deterministic = [lay.deterministic for lay in int_layers]
    for lay in int_layers:
        lay.deterministic = True
    try:
        yield
    finally:
        # sample and reset the previous behavior
        parameter.value  # pylint: disable=pointless-statement
        for lay, det in zip(int_layers, deterministic):
            lay.deterministic = det


def _fully_bounded_layers(data: pdata.Data) -> tp.List[_datalayers.BoundLayer]:
    """Extract fully bounded layers of a Data parameter"""
    layers = _datalayers.BoundLayer.filter_from(data)  # find bound layers
    layers = [  # keep only fully bounded layers
        lay for lay in layers if not any(b is None for b in lay.bounds)
    ]
    return layers


class Normalizer:
    """Hacky way to sample in the space defined by the parametrization.
    Given an vector of values between 0 and 1,
    the transform method samples in the bounds if provided,
    or using the provided function otherwise.
    This is used for samplers.
    Code of parametrization and/or this helper should definitely be
    updated to make it simpler and more robust
    """

    def __init__(
        self,
        reference: core.Parameter,
        unbounded_transform: tp.Optional[trans.Transform] = None,
        only_sampling: bool = False,
    ) -> None:
        self.reference = reference.spawn_child()
        self.reference.freeze()
        # initial check
        parameter = self.reference.spawn_child()
        parameter.set_standardized_data(np.linspace(-1, 1, self.reference.dimension))
        expected = parameter.get_standardized_data(reference=self.reference)
        self._ref_arrays = list_data(self.reference)
        arrays = list_data(parameter)
        check = np.concatenate(
            [x.get_standardized_data(reference=y) for x, y in zip(arrays, self._ref_arrays)], axis=0
        )
        self.working = True
        if not np.allclose(check, expected):
            self.working = False
            self._warn()
        self._only_sampling = only_sampling
        self.unbounded_transform = (
            trans.ArctanBound(0, 1) if unbounded_transform is None else unbounded_transform
        )
        self.fully_bounded = all(bool(_fully_bounded_layers(data)) for data in self._ref_arrays)

    def _warn(self) -> None:
        warnings.warn(
            f"Failed to find bounds for {self.reference}, quasi-random optimizer may be inefficient.\n"
            "Please open an issue on Nevergrad github"
        )

    def backward(self, x: tp.ArrayLike) -> np.ndarray:
        """Transform from [0, 1] to standardized space"""
        return self._apply(x, forward=False)

    def forward(self, x: tp.ArrayLike) -> np.ndarray:
        """Transform from standardized space to [0, 1]"""
        return self._apply(x, forward=True)

    def _apply(self, x: tp.ArrayLike, forward: bool = True) -> np.ndarray:
        utrans = self.unbounded_transform.forward if forward else self.unbounded_transform.backward
        y = np.array(x, copy=True, dtype=float)
        if not self.working:
            return utrans(y)
        try:
            self._apply_unsafe(y, forward=forward)  # inplace
        except Exception:  # pylint: disable=broad-except
            self._warn()
            y = utrans(y)
        return y

    def _apply_unsafe(self, x: np.ndarray, forward: bool = True) -> None:
        # modifies x in place
        start = 0
        utrans = self.unbounded_transform.forward if forward else self.unbounded_transform.backward
        for ref in self._ref_arrays:
            end = start + ref.dimension
            layers = _fully_bounded_layers(ref)
            if self._only_sampling:  # for samplers
                layers = [lay for lay in layers if lay.uniform_sampling]
            if not layers:
                x[start:end] = utrans(x[start:end])
            else:
                layer_index = layers[-1]._layer_index
                array = ref.spawn_child()
                if forward:
                    array.set_standardized_data(x[start:end])
                    x[start:end] = array._layers[layer_index].get_normalized_value()[:]  # type: ignore

                else:
                    normalized = x[start:end].reshape(ref._value.shape)
                    array._layers[layer_index].set_normalized_value(normalized)  # type: ignore
                    x[start:end] = array.get_standardized_data(reference=ref)
            start = end



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .core import Ant as Ant
from .core import Swimmer as Swimmer
from .core import HalfCheetah as HalfCheetah
from .core import Hopper as Hopper
from .core import Walker2d as Walker2d
from .core import Humanoid as Humanoid



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import time
from pathlib import Path
import logging
import os
import numpy as np
import nevergrad as ng
import nevergrad.common.typing as tp
from . import optimizerlib
from . import callbacks


# pylint: disable=unused-argument
def _func(x: tp.Any, y: tp.Any, blublu: str, array: tp.Any, multiobjective: bool = False) -> tp.Loss:
    return 12 if not multiobjective else [12, 12]


def test_log_parameters(tmp_path: Path) -> None:
    filepath = tmp_path / "logs.txt"
    cases = [0, np.int_(1), np.float_(2.0), np.nan, float("inf"), np.inf]
    instrum = ng.p.Instrumentation(
        ng.ops.mutations.Translation()(ng.p.Array(shape=(1,))),
        ng.p.Scalar(),
        blublu=ng.p.Choice(cases),
        array=ng.p.Array(shape=(3, 2)),
    )
    optimizer = optimizerlib.NoisyOnePlusOne(parametrization=instrum, budget=32)
    optimizer.register_callback("tell", ng.callbacks.ParametersLogger(filepath, append=False))
    optimizer.minimize(_func, verbosity=2)
    # pickling
    logger = callbacks.ParametersLogger(filepath)
    logs = logger.load_flattened()
    assert len(logs) == 32
    assert isinstance(logs[-1]["1"], float)
    assert len(logs[-1]) == 38
    logs = logger.load_flattened(max_list_elements=2)
    assert len(logs[-1]) == 34
    # deletion
    logger = callbacks.ParametersLogger(filepath, append=False)
    assert not logger.load()


def test_multiobjective_log_parameters(tmp_path: Path) -> None:
    filepath = tmp_path / "logs.txt"
    instrum = ng.p.Instrumentation(
        None, 2.0, blublu="blublu", array=ng.p.Array(shape=(3, 2)), multiobjective=True
    )
    optimizer = optimizerlib.OnePlusOne(parametrization=instrum, budget=2)
    optimizer.register_callback("tell", ng.callbacks.ParametersLogger(filepath, append=False))
    optimizer.minimize(_func, verbosity=2)
    # pickling
    logger = callbacks.ParametersLogger(filepath)
    logs = logger.load_flattened()
    assert len(logs) == 2


def test_chaining_log_parameters(tmp_path: Path) -> None:
    filepath = tmp_path / "logs.txt"
    params = ng.p.Instrumentation(
        None, 2.0, blublu="blublu", array=ng.p.Array(shape=(3, 2)), multiobjective=False
    )
    zmethods = ["CauchyLHSSearch", "DE", "CMA"]
    ztmp1 = [ng.optimizers.registry[zmet] for zmet in zmethods]
    optmodel = ng.families.Chaining(ztmp1, [50, 50])  #
    optim = optmodel(parametrization=params, budget=100, num_workers=3)
    logger = ng.callbacks.ParametersLogger(filepath)
    optim.register_callback("tell", logger)
    optim.minimize(_func, verbosity=2)
    # read
    logger = callbacks.ParametersLogger(filepath)
    logs = logger.load_flattened()
    assert len(logs) == 100


def test_dump_callback(tmp_path: Path) -> None:
    filepath = tmp_path / "pickle.pkl"
    optimizer = optimizerlib.OnePlusOne(parametrization=2, budget=32)
    optimizer.register_callback("tell", ng.callbacks.OptimizerDump(filepath))
    cand = optimizer.ask()
    assert not filepath.exists()
    optimizer.tell(cand, 0)
    assert filepath.exists()


def test_progressbar_dump(tmp_path: Path) -> None:
    filepath = tmp_path / "pickle.pkl"
    optimizer = optimizerlib.OnePlusOne(parametrization=2, budget=32)
    optimizer.register_callback("tell", ng.callbacks.ProgressBar())
    for _ in range(8):
        cand = optimizer.ask()
        optimizer.tell(cand, 0)
    optimizer.dump(filepath)
    # should keep working after dump
    cand = optimizer.ask()
    optimizer.tell(cand, 0)
    # and be reloadable
    optimizer = optimizerlib.OnePlusOne.load(filepath)
    for _ in range(12):
        cand = optimizer.ask()
        optimizer.tell(cand, 0)


class _EarlyStoppingTestee:
    def __init__(self, val=None, multi=False) -> None:
        self.num_calls = 0
        self.val = val
        self.multi = False

    def __call__(self, *args, **kwds) -> tp.Union[float, tp.Tuple]:
        self.num_calls += 1
        if self.val is not None:
            return self.val if not self.multi else (self.val, self.val)
        return np.random.rand() if not self.multi else (np.random.rand(), np.random.rand())


def test_early_stopping() -> None:
    instrum = ng.p.Instrumentation(None, 2.0, blublu="blublu", array=ng.p.Array(shape=(3, 2)))
    func = _EarlyStoppingTestee()
    optimizer = optimizerlib.OnePlusOne(parametrization=instrum, budget=100)
    early_stopping = ng.callbacks.EarlyStopping(lambda opt: opt.num_ask > 3)
    optimizer.register_callback("ask", early_stopping)
    optimizer.register_callback("ask", ng.callbacks.EarlyStopping.timer(100))  # should not get triggered
    optimizer.minimize(func, verbosity=2)
    # num_ask is set at the end of ask, so the callback sees the old value.
    assert func.num_calls == 4
    # below functions are included in the docstring of EarlyStopping
    assert optimizer.current_bests["minimum"].mean < 12
    assert optimizer.recommend().loss < 12  # type: ignore

    # test for no improvement
    func = _EarlyStoppingTestee(5)
    optimizer = optimizerlib.OnePlusOne(parametrization=instrum, budget=100)
    no_imp_window = 7
    optimizer.register_callback(
        "ask", ng.callbacks.EarlyStopping.no_improvement_stopper(no_imp_window)
    )  # should get triggered
    optimizer.minimize(func, verbosity=2)
    assert func.num_calls == no_imp_window + 2

    # test for no improvement multi objective
    func = _EarlyStoppingTestee(5, multi=True)
    optimizer = optimizerlib.OnePlusOne(parametrization=instrum, budget=100)
    no_imp_window = 5
    optimizer.register_callback(
        "ask", ng.callbacks.EarlyStopping.no_improvement_stopper(no_imp_window)
    )  # should get triggered
    optimizer.minimize(func, verbosity=2)
    assert func.num_calls == no_imp_window + 2


def test_duration_criterion() -> None:
    optim = optimizerlib.OnePlusOne(2, budget=100)
    crit = ng.callbacks._DurationCriterion(0.01)
    assert not crit(optim)
    assert not crit(optim)
    assert not crit(optim)
    time.sleep(0.01)
    assert crit(optim)


def test_optimization_logger(caplog) -> None:
    instrum = ng.p.Instrumentation(
        None, 2.0, blublu="blublu", array=ng.p.Array(shape=(3, 2)), multiobjective=False
    )
    logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))
    logger = logging.getLogger(__name__)
    optimizer = optimizerlib.OnePlusOne(parametrization=instrum, budget=3)
    optimizer.register_callback(
        "tell",
        callbacks.OptimizationLogger(
            logger=logger, log_level=logging.INFO, log_interval_tells=10, log_interval_seconds=0.1
        ),
    )
    with caplog.at_level(logging.INFO):
        optimizer.minimize(_func, verbosity=2)
    assert (
        "After 0, recommendation is Instrumentation(Tuple(None,2.0),Dict(array=Array{(3,2)},blublu=blublu,multiobjective=False))"
        in caplog.text
    )


def test_optimization_logger_MOO(caplog) -> None:
    instrum = ng.p.Instrumentation(
        None, 2.0, blublu="blublu", array=ng.p.Array(shape=(3, 2)), multiobjective=True
    )
    logging.basicConfig(level=os.environ.get("LOGLEVEL", "INFO"))
    logger = logging.getLogger(__name__)
    optimizer = optimizerlib.OnePlusOne(parametrization=instrum, budget=3)
    optimizer.register_callback(
        "tell",
        callbacks.OptimizationLogger(
            logger=logger, log_level=logging.INFO, log_interval_tells=10, log_interval_seconds=0.1
        ),
    )
    with caplog.at_level(logging.INFO):
        optimizer.minimize(_func, verbosity=2)
    assert (
        "After 0, the respective minimum loss for each objective in the pareto front is [12. 12.]"
        in caplog.text
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .oneshot import SamplingSearch
from .differentialevolution import DifferentialEvolution
from .optimizerlib import RandomSearchMaker, SQP, LHSSearch, DE, RandomSearch, MetaRecentering, MetaTuneRecentering  # type: ignore
from .optimizerlib import (
    ParametrizedMetaModel,
    ParametrizedOnePlusOne,
    ParametrizedCMA,
    ParametrizedBO,
    EMNA,
    CmaFmin2,
    NGOpt10,
    NGOpt12,
    BayesOptim,
    ConfPortfolio,
    DiagonalCMA,
    GeneticDE,
    TBPSA,
    NoisyOnePlusOne,
    RecombiningPortfolioOptimisticNoisyDiscreteOnePlusOne,
    OptimisticNoisyOnePlusOne,
)
from . import optimizerlib as opts
from .optimizerlib import CMA, Chaining, PSO, BO

# DE
OnePointDE = DifferentialEvolution(crossover="onepoint").set_name("OnePointDE", register=True)
ParametrizationDE = DifferentialEvolution(crossover="parametrization").set_name(
    "ParametrizationDE", register=True
)
MiniDE = DifferentialEvolution(initialization="gaussian", scale="mini").set_name("MiniDE", register=True)
MiniLhsDE = DifferentialEvolution(initialization="LHS", scale="mini").set_name("MiniLhsDE", register=True)
MiniQrDE = DifferentialEvolution(initialization="QR", scale="mini").set_name("MiniQrDE", register=True)
AlmostRotationInvariantDEAndBigPop = DifferentialEvolution(crossover=0.9, popsize="dimension").set_name(
    "AlmostRotationInvariantDEAndBigPop", register=True
)
BPRotationInvariantDE = DifferentialEvolution(crossover=1.0, popsize="large").set_name(
    "BPRotationInvariantDE", register=True
)

# CMA
MilliCMA = ParametrizedCMA(scale=1e-3).set_name("MilliCMA", register=True)
MicroCMA = ParametrizedCMA(scale=1e-6).set_name("MicroCMA", register=True)
FCMAs03 = ParametrizedCMA(fcmaes=True, scale=0.3).set_name("FCMAs03", register=True)
FCMAp13 = ParametrizedCMA(fcmaes=True, scale=0.1, popsize=13).set_name("FCMAp13", register=True)
ECMA = ParametrizedCMA(elitist=True).set_name("ECMA", register=True)
MetaModelDiagonalCMA = ParametrizedMetaModel(multivariate_optimizer=ParametrizedCMA(diagonal=True)).set_name(
    "MetaModelDiagonalCMA", register=True
)
MetaModelFmin2 = ParametrizedMetaModel(multivariate_optimizer=CmaFmin2).set_name(
    "MetaModelFmin2", register=True
)
MetaModelFmin2.no_parallelization = True
LSCMA = ParametrizedCMA(high_speed=False).set_name("LSCMA", register=True)
HSCMA = ParametrizedCMA(high_speed=True).set_name("HSCMA", register=True)
HSNeuralCMA = ParametrizedCMA(high_speed=True, algorithm="neural").set_name("HSNeuralCMA", register=True)
HSSVMCMA = ParametrizedCMA(high_speed=True, algorithm="svm").set_name("HSSVMCMA", register=True)
HSRFCMA = ParametrizedCMA(high_speed=True, algorithm="rf").set_name("HSRFCMA", register=True)
HSMetaModel = ParametrizedMetaModel(multivariate_optimizer=HSCMA).set_name("HSMetaModel", register=True)

# OnePlusOne
FastGADiscreteOnePlusOne = ParametrizedOnePlusOne(mutation="fastga").set_name(
    "FastGADiscreteOnePlusOne", register=True
)
DoubleFastGAOptimisticNoisyDiscreteOnePlusOne = ParametrizedOnePlusOne(
    noise_handling="optimistic", mutation="doublefastga"
).set_name("DoubleFastGAOptimisticNoisyDiscreteOnePlusOne", register=True)
RecombiningGA = ParametrizedOnePlusOne(
    mutation="doublefastga",
    crossover=True,
).set_name("RecombiningGA", register=True)
RotatedRecombiningGA = ParametrizedOnePlusOne(
    mutation="doublefastga", crossover=True, rotation=True
).set_name("RotatedRecombiningGA", register=True)
FastGAOptimisticNoisyDiscreteOnePlusOne = ParametrizedOnePlusOne(
    noise_handling="optimistic", mutation="fastga"
).set_name("FastGAOptimisticNoisyDiscreteOnePlusOne", register=True)
FastGANoisyDiscreteOnePlusOne = ParametrizedOnePlusOne(noise_handling="random", mutation="fastga").set_name(
    "FastGANoisyDiscreteOnePlusOne", register=True
)
PortfolioOptimisticNoisyDiscreteOnePlusOne = ParametrizedOnePlusOne(
    noise_handling="optimistic", mutation="portfolio"
).set_name("PortfolioOptimisticNoisyDiscreteOnePlusOne", register=True)
PortfolioNoisyDiscreteOnePlusOne = ParametrizedOnePlusOne(
    noise_handling="random", mutation="portfolio"
).set_name("PortfolioNoisyDiscreteOnePlusOne", register=True)
RecombiningOptimisticNoisyDiscreteOnePlusOne = ParametrizedOnePlusOne(
    crossover=True, mutation="discrete", noise_handling="optimistic"
).set_name("RecombiningOptimisticNoisyDiscreteOnePlusOne", register=True)

# BO
RBO = ParametrizedBO(initialization="random").set_name("RBO", register=True)
QRBO = ParametrizedBO(initialization="Hammersley").set_name("QRBO", register=True)
MidQRBO = ParametrizedBO(initialization="Hammersley", middle_point=True).set_name("MidQRBO", register=True)
LBO = ParametrizedBO(initialization="LHS").set_name("LBO", register=True)

# EMNA
IsoEMNA = EMNA(naive=False).set_name("IsoEMNA", register=True)
NaiveAnisoEMNA = EMNA(isotropic=False).set_name("NaiveAnisoEMNA", register=True)
AnisoEMNA = EMNA(naive=False, isotropic=False).set_name("AnisoEMNA", register=True)
IsoEMNATBPSA = EMNA(naive=False, population_size_adaptation=True).set_name("IsoEMNATBPSA", register=True)
NaiveIsoEMNATBPSA = EMNA(population_size_adaptation=True).set_name("NaiveIsoEMNATBPSA", register=True)
AnisoEMNATBPSA = EMNA(naive=False, isotropic=False, population_size_adaptation=True).set_name(
    "AnisoEMNATBPSA", register=True
)
NaiveAnisoEMNATBPSA = EMNA(isotropic=False, population_size_adaptation=True).set_name(
    "NaiveAnisoEMNATBPSA", register=True
)

# Recentering
MetaCauchyRecentering = SamplingSearch(
    cauchy=True, autorescale=True, sampler="Hammersley", scrambled=True
).set_name("MetaCauchyRecentering", register=True)
# Chaining
ChainCMASQP = Chaining([CMA, SQP], ["half"]).set_name("ChainCMASQP", register=True)
ChainCMASQP.no_parallelization = True
ChainDEwithR = Chaining([RandomSearch, DE], ["num_workers"]).set_name("ChainDEwithR", register=True)
ChainDEwithRsqrt = Chaining([RandomSearch, DE], ["sqrt"]).set_name("ChainDEwithRsqrt", register=True)
ChainDEwithRdim = Chaining([RandomSearch, DE], ["dimension"]).set_name("ChainDEwithRdim", register=True)
ChainDEwithR30 = Chaining([RandomSearch, DE], [30]).set_name("ChainDEwithR30", register=True)
ChainDEwithLHS = Chaining([LHSSearch, DE], ["num_workers"]).set_name("ChainDEwithLHS", register=True)
ChainDEwithLHSsqrt = Chaining([LHSSearch, DE], ["sqrt"]).set_name("ChainDEwithLHSsqrt", register=True)
ChainDEwithLHSdim = Chaining([LHSSearch, DE], ["dimension"]).set_name("ChainDEwithLHSdim", register=True)
ChainDEwithLHS30 = Chaining([LHSSearch, DE], [30]).set_name("ChainDEwithLHS30", register=True)
ChainDEwithMetaRecentering = Chaining([MetaRecentering, DE], ["num_workers"]).set_name(
    "ChainDEwithMetaRecentering", register=True
)
ChainDEwithMetaRecenteringsqrt = Chaining([MetaRecentering, DE], ["sqrt"]).set_name(
    "ChainDEwithMetaRecenteringsqrt", register=True
)
ChainDEwithMetaRecenteringdim = Chaining([MetaRecentering, DE], ["dimension"]).set_name(
    "ChainDEwithMetaRecenteringdim", register=True
)
ChainDEwithMetaRecentering30 = Chaining([MetaRecentering, DE], [30]).set_name(
    "ChainDEwithMetaRecentering30", register=True
)
ChainBOwithMetaTuneRecentering = Chaining([MetaTuneRecentering, BO], ["num_workers"]).set_name(
    "ChainBOwithMetaTuneRecentering", register=True
)
ChainBOwithMetaTuneRecenteringsqrt = Chaining([MetaTuneRecentering, BO], ["sqrt"]).set_name(
    "ChainBOwithMetaTuneRecenteringsqrt", register=True
)
ChainBOwithMetaTuneRecenteringdim = Chaining([MetaTuneRecentering, BO], ["dimension"]).set_name(
    "ChainBOwithMetaTuneRecenteringdim", register=True
)
ChainBOwithMetaTuneRecentering30 = Chaining([MetaTuneRecentering, BO], [30]).set_name(
    "ChainBOwithMetaTuneRecentering30", register=True
)

ChainDEwithMetaTuneRecentering = Chaining([MetaTuneRecentering, DE], ["num_workers"]).set_name(
    "ChainDEwithMetaTuneRecentering", register=True
)
ChainDEwithMetaTuneRecenteringsqrt = Chaining([MetaTuneRecentering, DE], ["sqrt"]).set_name(
    "ChainDEwithMetaTuneRecenteringsqrt", register=True
)
ChainDEwithMetaTuneRecenteringdim = Chaining([MetaTuneRecentering, DE], ["dimension"]).set_name(
    "ChainDEwithMetaTuneRecenteringdim", register=True
)
ChainDEwithMetaTuneRecentering30 = Chaining([MetaTuneRecentering, DE], [30]).set_name(
    "ChainDEwithMetaTuneRecentering30", register=True
)


ChainBOwithR = Chaining([RandomSearch, BO], ["num_workers"]).set_name("ChainBOwithR", register=True)
ChainBOwithRsqrt = Chaining([RandomSearch, BO], ["sqrt"]).set_name("ChainBOwithRsqrt", register=True)
ChainBOwithRdim = Chaining([RandomSearch, BO], ["dimension"]).set_name("ChainBOwithRdim", register=True)
ChainBOwithR30 = Chaining([RandomSearch, BO], [30]).set_name("ChainBOwithR30", register=True)
ChainBOwithLHS30 = Chaining([LHSSearch, BO], [30]).set_name("ChainBOwithLHS30", register=True)
ChainBOwithLHSsqrt = Chaining([LHSSearch, BO], ["sqrt"]).set_name("ChainBOwithLHSsqrt", register=True)
ChainBOwithLHSdim = Chaining([LHSSearch, BO], ["dimension"]).set_name("ChainBOwithLHSdim", register=True)
ChainBOwithLHS = Chaining([LHSSearch, BO], ["num_workers"]).set_name("ChainBOwithLHS", register=True)
ChainBOwithMetaRecentering30 = Chaining([MetaRecentering, BO], [30]).set_name(
    "ChainBOwithMetaRecentering30", register=True
)
ChainBOwithMetaRecenteringsqrt = Chaining([MetaRecentering, BO], ["sqrt"]).set_name(
    "ChainBOwithMetaRecenteringsqrt", register=True
)
ChainBOwithMetaRecenteringdim = Chaining([MetaRecentering, BO], ["dimension"]).set_name(
    "ChainBOwithMetaRecenteringdim", register=True
)
ChainBOwithMetaRecentering = Chaining([MetaRecentering, BO], ["num_workers"]).set_name(
    "ChainBOwithMetaRecentering", register=True
)

ChainPSOwithR = Chaining([RandomSearch, PSO], ["num_workers"]).set_name("ChainPSOwithR", register=True)
ChainPSOwithRsqrt = Chaining([RandomSearch, PSO], ["sqrt"]).set_name("ChainPSOwithRsqrt", register=True)
ChainPSOwithRdim = Chaining([RandomSearch, PSO], ["dimension"]).set_name("ChainPSOwithRdim", register=True)
ChainPSOwithR30 = Chaining([RandomSearch, PSO], [30]).set_name("ChainPSOwithR30", register=True)
ChainPSOwithLHS30 = Chaining([LHSSearch, PSO], [30]).set_name("ChainPSOwithLHS30", register=True)
ChainPSOwithLHSsqrt = Chaining([LHSSearch, PSO], ["sqrt"]).set_name("ChainPSOwithLHSsqrt", register=True)
ChainPSOwithLHSdim = Chaining([LHSSearch, PSO], ["dimension"]).set_name("ChainPSOwithLHSdim", register=True)
ChainPSOwithLHS = Chaining([LHSSearch, PSO], ["num_workers"]).set_name("ChainPSOwithLHS", register=True)
ChainPSOwithMetaRecentering30 = Chaining([MetaRecentering, PSO], [30]).set_name(
    "ChainPSOwithMetaRecentering30", register=True
)
ChainPSOwithMetaRecenteringsqrt = Chaining([MetaRecentering, PSO], ["sqrt"]).set_name(
    "ChainPSOwithMetaRecenteringsqrt", register=True
)
ChainPSOwithMetaRecenteringdim = Chaining([MetaRecentering, PSO], ["dimension"]).set_name(
    "ChainPSOwithMetaRecenteringdim", register=True
)
ChainPSOwithMetaRecentering = Chaining([MetaRecentering, PSO], ["num_workers"]).set_name(
    "ChainPSOwithMetaRecentering", register=True
)

ChainCMAwithR = Chaining([RandomSearch, CMA], ["num_workers"]).set_name("ChainCMAwithR", register=True)
ChainCMAwithRsqrt = Chaining([RandomSearch, CMA], ["sqrt"]).set_name("ChainCMAwithRsqrt", register=True)
ChainCMAwithRdim = Chaining([RandomSearch, CMA], ["dimension"]).set_name("ChainCMAwithRdim", register=True)
ChainCMAwithR30 = Chaining([RandomSearch, CMA], [30]).set_name("ChainCMAwithR30", register=True)
ChainCMAwithLHS30 = Chaining([LHSSearch, CMA], [30]).set_name("ChainCMAwithLHS30", register=True)
ChainCMAwithLHSsqrt = Chaining([LHSSearch, CMA], ["sqrt"]).set_name("ChainCMAwithLHSsqrt", register=True)
ChainCMAwithLHSdim = Chaining([LHSSearch, CMA], ["dimension"]).set_name("ChainCMAwithLHSdim", register=True)
ChainCMAwithLHS = Chaining([LHSSearch, CMA], ["num_workers"]).set_name("ChainCMAwithLHS", register=True)
ChainCMAwithMetaRecentering30 = Chaining([MetaRecentering, CMA], [30]).set_name(
    "ChainCMAwithMetaRecentering30", register=True
)
ChainCMAwithMetaRecenteringsqrt = Chaining([MetaRecentering, CMA], ["sqrt"]).set_name(
    "ChainCMAwithMetaRecenteringsqrt", register=True
)
ChainCMAwithMetaRecenteringdim = Chaining([MetaRecentering, CMA], ["dimension"]).set_name(
    "ChainCMAwithMetaRecenteringdim", register=True
)
ChainCMAwithMetaRecentering = Chaining([MetaRecentering, CMA], ["num_workers"]).set_name(
    "ChainCMAwithMetaRecentering", register=True
)

# Random search
Zero = RandomSearchMaker(scale=0.0).set_name("Zero", register=True)
StupidRandom = RandomSearchMaker(stupid=True).set_name("StupidRandom", register=True)
CauchyRandomSearch = RandomSearchMaker(sampler="cauchy").set_name("CauchyRandomSearch", register=True)
RandomScaleRandomSearch = RandomSearchMaker(scale="random", middle_point=True).set_name(
    "RandomScaleRandomSearch", register=True
)
RandomScaleRandomSearchPlusMiddlePoint = RandomSearchMaker(scale="random", middle_point=True).set_name(
    "RandomScaleRandomSearchPlusMiddlePoint", register=True
)

# quasi-random
RescaleScrHammersleySearch = SamplingSearch(sampler="Hammersley", scrambled=True, rescaled=True).set_name(
    "RescaleScrHammersleySearch", register=True
)
AvgHammersleySearch = SamplingSearch(sampler="Hammersley", recommendation_rule="average_of_best").set_name(
    "AvgHammersleySearch", register=True
)
AvgHammersleySearchPlusMiddlePoint = SamplingSearch(
    sampler="Hammersley", middle_point=True, recommendation_rule="average_of_best"
).set_name("AvgHammersleySearchPlusMiddlePoint", register=True)
HullCenterHullAvgRandomSearch = SamplingSearch(
    sampler="Random", recommendation_rule="average_of_hull_best"
).set_name("HullCenterHullAvgRandomSearch", register=True)
AvgRandomSearch = SamplingSearch(sampler="Random", recommendation_rule="average_of_best").set_name(
    "AvgRandomSearch", register=True
)

# Recommendation rule = average of k best, with k = n / 2^d.
TEAvgScrHammersleySearchPlusMiddlePoint = SamplingSearch(
    scrambled=True, sampler="Hammersley", middle_point=True, recommendation_rule="average_of_exp_best"
).set_name("TEAvgScrHammersleySearchPlusMiddlePoint", register=True)
TEAvgScrHammersleySearch = SamplingSearch(
    sampler="Hammersley", scrambled=True, recommendation_rule="average_of_exp_best"
).set_name("TEAvgScrHammersleySearch", register=True)
TEAvgRandomSearch = SamplingSearch(sampler="Random", recommendation_rule="average_of_exp_best").set_name(
    "TEAvgRandomSearch", register=True
)
TEAvgCauchyScrHammersleySearch = SamplingSearch(
    cauchy=True, sampler="Hammersley", scrambled=True, recommendation_rule="average_of_exp_best"
).set_name("TEAvgCauchyScrHammersleySearch", register=True)
TEAvgLHSSearch = SamplingSearch(sampler="LHS", recommendation_rule="average_of_exp_best").set_name(
    "TEAvgLHSSearch", register=True
)
TEAvgCauchyLHSSearch = SamplingSearch(
    sampler="LHS", cauchy=True, recommendation_rule="average_of_exp_best"
).set_name("TEAvgCauchyLHSSearch", register=True)

# Recommendation rule = by convex hull.
HullCenterHullAvgScrHaltonSearch = SamplingSearch(
    scrambled=True, recommendation_rule="average_of_hull_best"
).set_name("HullCenterHullAvgScrHaltonSearch", register=True)
HullCenterHullAvgScrHaltonSearchPlusMiddlePoint = SamplingSearch(
    middle_point=True, scrambled=True, recommendation_rule="average_of_hull_best"
).set_name("HullCenterHullAvgScrHaltonSearchPlusMiddlePoint", register=True)
HullCenterHullAvgScrHammersleySearchPlusMiddlePoint = SamplingSearch(
    scrambled=True, sampler="Hammersley", middle_point=True, recommendation_rule="average_of_hull_best"
).set_name("HullCenterHullAvgScrHammersleySearchPlusMiddlePoint", register=True)
HullCenterHullAvgLargeHammersleySearch = SamplingSearch(
    scale=100.0, sampler="Hammersley", recommendation_rule="average_of_hull_best"
).set_name("HullCenterHullAvgLargeHammersleySearch", register=True)
HullCenterHullAvgScrHammersleySearch = SamplingSearch(
    sampler="Hammersley", scrambled=True, recommendation_rule="average_of_hull_best"
).set_name("HullCenterHullAvgScrHammersleySearch", register=True)
HullCenterHullAvgCauchyScrHammersleySearch = SamplingSearch(
    cauchy=True, sampler="Hammersley", scrambled=True, recommendation_rule="average_of_hull_best"
).set_name("HullCenterHullAvgCauchyScrHammersleySearch", register=True)
HullCenterHullAvgLHSSearch = SamplingSearch(
    sampler="LHS", recommendation_rule="average_of_hull_best"
).set_name("HullCenterHullAvgLHSSearch", register=True)
HullCenterHullAvgCauchyLHSSearch = SamplingSearch(
    sampler="LHS", cauchy=True, recommendation_rule="average_of_hull_best"
).set_name("HullCenterHullAvgCauchyLHSSearch", register=True)

# Split on top of competence map.
MetaNGOpt10 = opts.ConfSplitOptimizer(
    multivariate_optimizer=NGOpt10, monovariate_optimizer=NGOpt10, non_deterministic_descriptor=False
).set_name("MetaNGOpt10", register=True)

# Multiple single runs for multi-objective optimization.
NGOptSingle9 = opts.MultipleSingleRuns(num_single_runs=9, base_optimizer=NGOpt12).set_name(
    "NGOptSingle9", register=True
)
NGOptSingle16 = opts.MultipleSingleRuns(num_single_runs=16, base_optimizer=NGOpt12).set_name(
    "NGOptSingle16", register=True
)
NGOptSingle25 = opts.MultipleSingleRuns(num_single_runs=25, base_optimizer=NGOpt12).set_name(
    "NGOptSingle25", register=True
)

# noisy splitters
Noisy13Splits = opts.NoisySplit(num_optims=13, discrete=False).set_name("Noisy13Splits", register=True)
NoisyInfSplits = opts.NoisySplit(num_optims=float("inf"), discrete=False).set_name(
    "NoisyInfSplits", register=True
)
DiscreteNoisy13Splits = opts.NoisySplit(num_optims=13, discrete=True).set_name(
    "DiscreteNoisy13Splits", register=True
)
DiscreteNoisyInfSplits = opts.NoisySplit(num_optims=float("inf"), discrete=True).set_name(
    "DiscreteNoisyInfSplits", register=True
)

# PCA-BO
# Testing the influence of n_components on the performance of PCABO
PCABO80 = BayesOptim(pca=True, n_components=0.80).set_name("PCABO80", register=True)

# Testing the influence of the DoE size on the performance of PCABO
PCABO95DoE20 = BayesOptim(pca=True, n_components=0.95, prop_doe_factor=0.20).set_name(
    "PCABO95DoE20", register=True
)
SparseDiscreteOnePlusOne = ParametrizedOnePlusOne(mutation="discrete", sparse=True).set_name(
    "SparseDiscreteOnePlusOne", register=True
)

# Specifically for RL.
MixDeterministicRL = ConfPortfolio(optimizers=[DiagonalCMA, PSO, GeneticDE]).set_name(
    "MixDeterministicRL", register=True
)
SpecialRL = Chaining([MixDeterministicRL, TBPSA], ["half"]).set_name("SpecialRL", register=True)
NoisyRL1 = Chaining([MixDeterministicRL, NoisyOnePlusOne], ["half"]).set_name("NoisyRL1", register=True)
NoisyRL2 = Chaining(
    [MixDeterministicRL, RecombiningPortfolioOptimisticNoisyDiscreteOnePlusOne], ["half"]
).set_name("NoisyRL2", register=True)
NoisyRL3 = Chaining([MixDeterministicRL, OptimisticNoisyOnePlusOne], ["half"]).set_name(
    "NoisyRL3", register=True
)

# High-Speed variants
HSDE = DifferentialEvolution(high_speed=True).set_name("HSDE", register=True)
LhsHSDE = DifferentialEvolution(initialization="LHS", high_speed=True).set_name("LhsHSDE", register=True)



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from abc import ABC, abstractmethod
import math


class teampursuit(ABC):

    # constants
    friction_coefficient = 0.0025
    drafting_coefficients = [0.75, 0.65, 0.55]
    gravitational_acceleration = 9.80665
    time_step = 0.1

    transition_time = 0.12

    relative_humidity = 0.5

    temperature = 20.0
    barometric_pressure = 1013.25

    air_density = None
    team = None

    def __init__(self):
        self.update_air_density()

    def set_temperature(self, temperature):
        if temperature < 0.0 or temperature > 40:
            raise ValueError("Temperature must be in range 0-40C")
        else:
            self.temperature = temperature
            self.update_air_density()

    def set_barometric_pressure(self, barometric_pressure):
        if barometric_pressure < 800.0 or barometric_pressure > 1200.0:
            raise ValueError("Barometric pressure must be in the range 800-1200 hPa")
        else:
            self.barometric_pressure = barometric_pressure
            self.update_air_density()

    def set_relative_humidity(self, relative_humidity):
        if relative_humidity < 0.0 or relative_humidity > 1.0:
            raise ValueError("Relative humidity must be in the range 0-1")
        else:
            self.relative_humidity = relative_humidity
            self.update_air_density()

    def set_height(self, cyclist_id, height):
        if cyclist_id >= len(self.team):
            raise ValueError(f"Cyclist identifier must be in the range 0-{len(self.team)}")
        else:
            self.team[cyclist_id].set_height(height)

    def set_weight(self, cyclist_id, weight):
        if cyclist_id >= len(self.team):
            raise ValueError(f"Cyclist identifier must be in the range 0-{len(self.team)}")
        else:
            self.team[cyclist_id].set_weight(weight)

    def set_mean_maximum_power(self, cyclist_id, mean_maximum_power):
        if cyclist_id >= len(self.team):
            raise ValueError(f"Cyclist identifier must be in the range 0-{len(self.team)}")
        else:
            self.team[cyclist_id].set_mean_maximum_power(mean_maximum_power)

    def get_temperature(self):
        return self.temperature

    def get_barometric_pressure(self):
        return self.barometric_pressure

    def get_relative_humidity(self):
        return self.relative_humidity

    def get_height(self, cyclist_id):
        if cyclist_id >= len(self.team):
            raise ValueError(f"Cyclist identifier must be in the range 0-{len(self.team)}")
        else:
            return self.team[cyclist_id].get_height()

    def get_weight(self, cyclist_id):
        if cyclist_id >= len(self.team):
            raise ValueError(f"Cyclist identifier must be in the range 0-{len(self.team)}")
        else:
            return self.team[cyclist_id].get_weight()

    def get_mean_maximum_power(self, cyclist_id):
        if cyclist_id >= len(self.team):
            raise ValueError(f"Cyclist identifier must be in the range 0-{len(self.team)}")
        else:
            return self.team[cyclist_id].get_mean_maximum_power()

    @abstractmethod
    def simulate(self, transition_strategy, pacing_strategy):
        pass

    def update_air_density(self):
        pp_water_vapour = (
            100
            * self.relative_humidity
            * (
                6.1078
                * math.pow(
                    10,
                    (((7.5 * (self.temperature + 273.15)) - 2048.625)) / (self.temperature + 273.15 - 35.85),
                )
            )
        )
        pp_dry_air = 100 * self.barometric_pressure - pp_water_vapour
        self.air_density = (pp_dry_air / (287.058 * (self.temperature + 273.15))) + (
            pp_water_vapour / (461.495 * (self.temperature + 273.15))
        )

    def cyclists_remaining(self):
        cyclists_remaining = 0
        for i in range(0, len(self.team)):
            if self.team[i].get_remaining_energy() > 0.0:
                cyclists_remaining += 1
            else:
                self.team[i].set_position(0)
        return cyclists_remaining

    def leader(self):
        for i in range(0, len(self.team)):
            if (self.team[i].get_position()) == 1:
                return self.team[i]
        return None

    def transition(self):
        for i in range(0, len(self.team)):
            if self.team[i].get_position() > 0:
                if self.team[i].get_position() == 1:
                    self.team[i].set_position(self.cyclists_remaining())
                else:
                    self.team[i].set_position(self.team[i].get_position() - 1)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .core import UnitCommitmentProblem as UnitCommitmentProblem



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import uuid
import copy
import bisect
from enum import Enum
import numpy as np
from nevergrad.common import errors
import nevergrad.common.typing as tp


L = tp.TypeVar("L", bound="Layered")
F = tp.TypeVar("F", bound="Filterable")
In = tp.TypeVar("In")
Out = tp.TypeVar("Out")


class Level(Enum):
    """Lower level is deeper in the structure"""

    ROOT = 0
    OPERATION = 10

    # final
    INTEGER_CASTING = 800
    ARRAY_CASTING = 900
    SCALAR_CASTING = 950
    CONSTRAINT = 1000  # must be the last layer


class Layered:
    """Hidden API for overriding/modifying the behavior of a Parameter,
    which is itself a Layered object.

    Layers can be added and will be ordered depending on their level
    """

    _LAYER_LEVEL = Level.OPERATION  # this provides an order for the layers

    def __init__(self) -> None:
        self._layers = [self]
        self._layer_index = 0
        self._name: tp.Optional[str] = None
        self.uid = uuid.uuid4().hex

    def add_layer(self: L, other: "Layered") -> L:
        """Adds a layer which will modify the object behavior"""
        if self is not self._layers[0] or self._LAYER_LEVEL != Level.ROOT:
            raise errors.NevergradRuntimeError("Layers can only be added from the root.")
        if len(other._layers) > 1:
            raise errors.NevergradRuntimeError("Cannot append multiple layers at once")
        if other._LAYER_LEVEL.value >= self._layers[-1]._LAYER_LEVEL.value:
            other._layer_index = len(self._layers)
            self._layers.append(other)
        else:
            levels = [x._LAYER_LEVEL.value for x in self._layers]
            ind = bisect.bisect_right(levels, other._LAYER_LEVEL.value)
            self._layers.insert(ind, other)
            for k, x in enumerate(self._layers):
                x._layer_index = k
        other._layers = self._layers
        other._on_layer_added()
        return self

    def _on_layer_added(self) -> None:
        """Hook called when the layer is added to another one"""

    def _call_deeper(self, name: str, *args: tp.Any, **kwargs: tp.Any) -> tp.Any:
        if self._layers[self._layer_index] is not self:
            layers = [f"{lay.name}({lay._layer_index})" for lay in self._layers]
            raise errors.NevergradRuntimeError(
                "Layer indexing has changed for an unknown reason. Please open an issue:\n"
                f"Caller at index {self._layer_index}: {self.name}"
                f"Layers: {layers}.\n"
            )
        if not name.startswith("_layered_"):
            raise errors.NevergradValueError("For consistency, only _layered functions can be used.")
        for layer in reversed(self._layers[: self._layer_index]):
            func = getattr(layer, name)
            if func.__func__ is not getattr(Layered, name):  # skip unecessary stack calls
                return func(*args, **kwargs)
        types = [type(x) for x in self._layers]
        raise errors.NevergradNotImplementedError(f"No implementation for {name} on layers: {types}.")
        # ALTERNATIVE (stacking all calls):
        # if not self._layer_index:  # root must have an implementation
        #    raise errors.NevergradNotImplementedError
        # return getattr(self._layers[self._layer_index - 1], name)(*args, **kwargs)

    def _layered_get_value(self) -> tp.Any:
        return self._call_deeper("_layered_get_value")

    def _layered_set_value(self, value: tp.Any) -> tp.Any:
        return self._call_deeper("_layered_set_value", value)

    def _layered_del_value(self) -> None:
        pass  # called independently on each layer

    def _layered_sample(self) -> "Layered":
        return self._call_deeper("_layered_sample")  # type: ignore

    def _layered_mutate(self) -> None:
        self._call_deeper("_layered_mutate")

    def _layered_recombine(self, *args: "Layered") -> None:
        self._call_deeper("_layered_recombine", *args)

    @property
    def random_state(self) -> np.random.RandomState:
        return self._layers[0].random_state  # use the root random state

    def copy(self: L) -> L:
        """Creates a new unattached layer with the same behavior"""
        new = copy.copy(self)
        new._layers = [new]
        new._layer_index = 0
        if not self._layer_index:  # attach sublayers if root
            for layer in self._layers[1:]:
                new.add_layer(layer.copy())
        return new

    # naming capacity

    def _get_name(self) -> str:
        """Internal implementation of parameter name. This should be value independant, and should not account
        for internal/model parameters.
        """
        return self.__class__.__name__

    def __repr__(self) -> str:
        return self.name

    @property
    def name(self) -> str:
        """Name of the parameter
        This is used to keep track of how this Parameter is configured (included through internal/model parameters),
        mostly for reproducibility A default version is always provided, but can be overriden directly
        through the attribute, or through the set_name method (which allows chaining).
        """
        if self._name is not None:
            return self._name
        return self._get_name()

    @name.setter
    def name(self, name: str) -> None:
        self.set_name(name)  # with_name allows chaining

    def set_name(self: L, name: str) -> L:
        """Sets a name and return the current instrumentation (for chaining)

        Parameters
        ----------
        name: str
            new name to use to represent the Parameter
        """
        self._name = name
        return self


class ValueProperty(tp.Generic[In, Out]):
    """Typed property (descriptor) object so that the value attribute of
    Parameter objects fetches _layered_get_value and _layered_set_value methods
    """

    # This uses the descriptor protocol, like a property:
    # See https://docs.python.org/3/howto/descriptor.html
    #
    # Basically parameter.value calls parameter.value.__get__
    # and then parameter._layered_get_value
    def __init__(self) -> None:
        self.__doc__ = """Value of the Parameter, which should be sent to the function
        to optimize.

        Example
        -------
        >>> ng.p.Array(shape=(2,)).value
        array([0., 0.])
        """

    def __get__(self, obj: Layered, objtype: tp.Optional[tp.Type[object]] = None) -> Out:
        # pylint: disable=pointless-statement
        if hasattr(obj, "random_state"):
            getattr(obj, "random_state")  # very hacky way to initialize random_state
        return obj._layers[-1]._layered_get_value()  # type: ignore

    def __set__(self, obj: Layered, value: In) -> None:
        self.__delete__(obj)
        obj._layers[-1]._layered_set_value(value)

    def __delete__(self, obj: Layered) -> None:
        for layer in obj._layers:
            layer._layered_del_value()


# Basic data layers


class _ScalarCasting(Layered):
    """Cast Array as a scalar"""

    _LAYER_LEVEL = Level.SCALAR_CASTING

    def _layered_get_value(self) -> float:
        out = super()._layered_get_value()  # pulls from previous layer
        if not isinstance(out, np.ndarray) or not out.size == 1:
            raise errors.NevergradRuntimeError("Scalar casting can only be applied to size=1 Data parameters")
        integer = np.issubdtype(out.dtype, np.integer)
        out = (int if integer else float)(out[0])
        return out  # type: ignore

    def _layered_set_value(self, value: tp.Any) -> None:
        if not isinstance(value, (float, int, np.float_, np.int_)):
            raise TypeError(f"Received a {type(value)} in place of a scalar (float, int)")
        super()._layered_set_value(np.array([value], dtype=float))


class ArrayCasting(Layered):
    """Cast inputs of type tuple/list etc to array"""

    _LAYER_LEVEL = Level.ARRAY_CASTING

    def _layered_set_value(self, value: tp.ArrayLike) -> None:
        if not isinstance(value, (np.ndarray, tuple, list)):
            raise TypeError(f"Received a {type(value)} in place of a np.ndarray/tuple/list")
        super()._layered_set_value(np.asarray(value))


class Filterable:
    @classmethod
    def filter_from(cls: tp.Type[F], parameter: Layered) -> tp.List[F]:
        return [x for x in parameter._layers if isinstance(x, cls)]  # type: ignore


class Int(Layered, Filterable):
    """Cast Data as integer (or integer array)

    Parameters
    ----------
    deterministic: bool
        if True, the data is rounded to the closest integer, if False, both surrounded
        integers can be sampled inversely proportionally to how close the actual value
        is from the integers.

    Example
    -------
    0.2 is cast to 0 in deterministic mode, and either 0 (80% chance) or 1 (20% chance) in
    non-deterministic mode

    Usage
    -----
    :code:`param = ng.ops.Int(deterministic=True)(ng.p.Array(shape=(3,)))`
    """

    _LAYER_LEVEL = Level.INTEGER_CASTING

    def __init__(self, deterministic: bool = True) -> None:
        super().__init__()
        self.arity: tp.Optional[int] = None
        self.ordered = True
        self.deterministic = deterministic
        self._cache: tp.Optional[np.ndarray] = None

    def _get_name(self) -> str:
        tag = "" if self.deterministic else "{rand}"
        return self.__class__.__name__ + tag

    def _layered_get_value(self) -> np.ndarray:
        if self._cache is not None:
            return self._cache
        bounds = self._layers[0].bounds  # type: ignore
        out = super()._layered_get_value()
        if not self.deterministic:
            out += self.random_state.rand(*out.shape) - 0.5
        out = np.round(out).astype(int)
        # make sure rounding does not reach beyond the bounds
        eps = 1e-12
        if bounds[0] is not None:
            out = np.maximum(_to_int(bounds[0] + 0.5 - eps), out)
        if bounds[1] is not None:
            out = np.minimum(_to_int(bounds[1] - 0.5 + eps), out)
        # return out
        self._cache = out
        return self._cache

    def _layered_del_value(self) -> None:
        self._cache = None  # clear cache!

    def __call__(self, layered: L) -> L:
        """Creates a new Data instance with int-casting"""
        from . import data  # pylint: disable=import-outside-toplevel

        if not isinstance(layered, data.Data):
            raise ValueError("Only data parameters can use Int operation")
        out = layered.copy()
        out.add_layer(self)
        return out  # type: ignore


def _to_int(value: tp.Union[float, np.ndarray]) -> tp.Union[int, np.ndarray]:
    """Returns a int from a float, or a int typed array from a float array"""
    if not isinstance(value, np.ndarray):
        return int(np.round(value))
    else:
        return np.round(value).astype(int)



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math


class simulationresult:
    def __init__(self, finish_time, proportion_completed, energy_remaining, velocity_profile):
        self.finish_time = finish_time
        self.energy_remaining = energy_remaining
        self.proportion_completed = proportion_completed
        self.velocity_profile = velocity_profile
        self.results = []

    def get_finish_time(self):
        return self.finish_time

    def get_proportion_completed(self):
        return self.proportion_completed

    def get_energy_remaining(self):
        return self.energy_remaining

    def get_velocity_profile(self):
        return self.velocity_profile

    def to_string(self):
        output = "Simulation Result\n-----------------\n"
        if self.finish_time < math.inf:
            output = f"{output} Finish Time: {self.finish_time} seconds\n"
            for i in range(0, len(self.energy_remaining)):
                output = f"{output} Cyclist {i+1} Energy Remaining: {self.energy_remaining[i]} joules\n"
        else:
            output = f"{output} Riders had insufficient energy for race completion\nProportion of race completed: {self.proportion_completed * 100}%\n"
        return output



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import typing as tp
import numpy as np
from nevergrad.common import testing
from . import mutation
from .data import Array


@testing.parametrized(
    fft=(True, [3, 3, 5, 5]),
    real=(False, [4, 5, 5, 4]),
)
def test_crossover(fft: bool, expected: tp.List[int]) -> None:
    x1 = 4 * np.ones((2, 4))
    x2 = 5 * np.ones((2, 4)) if not fft else np.arange(8).reshape((2, 4))
    array = mutation.Crossover(axis=1, fft=fft)(Array(init=x1))
    co = array._layers[1]
    assert isinstance(co, mutation.Crossover)
    co.random_state.seed(12)
    out = co._apply_array((x1, x2))
    expected_arr = np.ones((2, 1)).dot([expected])
    np.testing.assert_array_equal(out, expected_arr)


def test_ravel_crossover() -> None:
    co = mutation.RavelCrossover()
    x1 = co(Array(init=5 * np.ones((2, 4))))
    x2 = co(Array(init=4 * np.ones((2, 4))))
    x2.random_state.seed(12)
    x2.recombine(x1)
    expected = [[4, 5, 5, 5], [4, 4, 4, 4]]
    np.testing.assert_array_equal(x2.value, expected)


def test_ravel_crossover_array() -> None:
    x1 = 4 * np.ones((2, 4))
    x2 = 5 * np.ones((2, 4))
    a1, a2 = (mutation.RavelCrossover()(Array(init=x)) for x in (x1, x2))
    a1.random_state.seed(12)
    a1.recombine(a2)
    expected = [[4, 5, 5, 5], [4, 4, 4, 4]]
    np.testing.assert_array_equal(a1.value, expected)


def test_local_gaussian() -> None:
    init = 4.0 * np.ones((2, 4))
    lg = mutation.LocalGaussian(axes=1, size=2)
    x = lg(Array(init=np.array(init)))
    x.random_state.seed(12)
    x.mutate()
    expected = np.ones((2, 1)).dot([[1, 0, 0, 1]])
    np.testing.assert_array_equal(x.value == init, expected)


def test_choice() -> None:
    lg = mutation.LocalGaussian(axes=1, size=2)
    jump = mutation.Jumping(axis=0, size=2)
    roll = mutation.Translation(0)
    Mut = mutation.MutationChoice([lg, jump, roll])
    x = Mut(Array(init=4.0 * np.ones((2, 4))))
    x.copy().mutate()


def test_cauchy() -> None:
    array = mutation.Cauchy()(Array(shape=(2, 4)))
    array.mutate()


# def test_proba_local_gaussian() -> None:
#     init = 4.0 * np.ones((2, 8))
#     x = Array(init=np.array(init))
#     lg = mutation.ProbaLocalGaussian(axis=1, shape=x.value.shape)
#     lg.parameters["ratio"].value = 0.3
#     pattern = [0, 0, 100, 100, 0, 0, 0, 0]
#     lg.parameters["positions"].value = pattern
#     lg.apply([x])
#     expected = np.ones((2, 1)).dot([pattern]) == 0
#     np.testing.assert_array_equal(x.value == init, expected)


def test_translation() -> None:
    x = np.arange(4)[:, None].dot(np.ones((1, 2)))
    roll = mutation.Translation(0)
    array = roll(Array(init=x))
    array.random_state.seed(12)
    array.mutate()
    expected = np.array([1, 2, 3, 0])[:, None].dot(np.ones((1, 2)))
    np.testing.assert_array_equal(array.value, expected)


def test_jump() -> None:
    x = np.arange(6)[:, None].dot(np.ones((1, 2)))
    jump = mutation.Jumping(axis=0, size=5)
    array = jump(Array(init=x))
    array.random_state.seed(38)
    array.mutate()
    expected = np.array([0, 3, 4, 1, 2, 5])[:, None].dot(np.ones((1, 2)))
    np.testing.assert_array_equal(array.value, expected)


# def test_tuned_translation() -> None:
#     x = np.arange(4)[:, None].dot(np.ones((1, 2)))
#     roll = mutation.TunedTranslation(0, shape=x.shape)
#     roll.random_state.seed(12)
#     out = roll._apply_array([x])
#     expected = np.array([3, 0, 1, 2])[:, None].dot(np.ones((1, 2)))
#     np.testing.assert_array_equal(out, expected)
#     roll.mutate()
#     assert np.sum(np.abs(roll.shift.indices._value)) > 0


@testing.parametrized(
    all_none=(None, None),
    d2=((1, 2), None),
    d1=((1), None),
)
def test_crossover_axis(axis: tp.Optional[tp.Tuple[int, ...]], max_size: tp.Optional[int]) -> None:
    shape = (6, 8, 10)
    co = mutation.Crossover(axis=axis, max_size=max_size)
    x1 = co(Array(init=4 * np.ones(shape)))
    x2 = co(Array(init=5 * np.ones(shape)))
    x1.random_state.seed(12)
    x1.recombine(x2)
    np.testing.assert_array_equal(
        x1.value.shape, shape
    )  # this basically only test that it did not raise an error


@testing.parametrized(
    w1=(1, [1, 2, 5, 9]),
    w2=(2, [3, 7, 14, 10]),
    w3=(3, [8, 16, 15, 12]),
    w4=(4, [17, 17, 17, 17]),
)
def test_rolling_mean(window: int, expected: tp.List[int]) -> None:
    output = mutation.rolling_mean(np.array([1, 2, 5, 9]), window)
    np.testing.assert_array_equal(output, expected)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# The source code in this module was created by Antoine Moreau
# from University Clermont Auvergne, CNRS, SIGMA Clermont, Institut Pascal.
#
# Publication:
# - Aliou Barry, Mamadou; Berthier, Vincent; Wilts, Bodo D.; Cambourieux, Marie-Claire; Pollès, Rémi;
#   Teytaud, Olivier; Centeno, Emmanuel; Biais, Nicolas; Moreau, Antoine (2018)
#   Evolutionary algorithms converge towards evolved biological photonic structures,
#   https://arxiv.org/abs/1808.04689
# - Defrance, J., Lemaître, C., Ajib, R., Benedicto, J., Mallet, E., Pollès, R., Plumey, J.-P.,
#   Mihailovic, M., Centeno, E., Ciracì, C., Smith, D.R. and Moreau, A., 2016.
#   Moosh: A Numerical Swiss Army Knife for the Optics of Multilayers in Octave/Matlab. Journal of Open Research Software, 4(1), p.e13.

import typing as tp
from pathlib import Path
import numpy as np
from scipy.linalg import toeplitz

# pylint: disable=blacklisted-name,too-many-locals,too-many-arguments


def bragg(X: np.ndarray) -> float:
    """
    Cost function for the Bragg mirror problem: maximizing the reflection
    when the refractive index are given for all the layers.
    Input: a vector whose components represent each the thickness of each
    layer.
    https://hal.archives-ouvertes.fr/hal-02613161
    """
    lam = 600
    bar = int(np.size(X) / 2)
    n = np.concatenate(([1], np.sqrt(X[0:bar]), [1.7320508075688772]))
    type_ = np.arange(0, bar + 2)
    hauteur = np.concatenate(([0], X[bar : 2 * bar], [0]))  # type: ignore
    tmp = np.tan(2 * np.pi * n[type_] * hauteur / lam)
    # Specific to this substrate.
    Z = n[-1]
    for k in range(np.size(type_) - 1, 0, -1):
        Z = (Z - 1j * n[type_[k]] * tmp[k]) / (1 - 1j * tmp[k] * Z / n[type_[k]])
    # Specific to air.
    r = (1 - Z) / (1 + Z)
    c = np.real(1 - r * np.conj(r))
    return float(c)


def chirped(X: np.ndarray) -> float:
    lam = np.linspace(500, 800, 50)
    n = np.array([1, 1.4142135623730951, 1.7320508075688772])
    type_ = np.concatenate(([0], np.tile([2, 1], int(np.size(X) / 2)), [2]))  # type: ignore
    hauteur = np.concatenate(([0], X, [0]))  # type: ignore
    r = np.zeros(np.size(lam)) + 0j
    for m in range(0, np.size(lam)):
        # Specific to this substrate.
        tmp = np.tan(2 * np.pi * n[type_] * hauteur / lam[m])
        Z = 1.7320508075688772
        for k in range(np.size(type_) - 1, 0, -1):
            Z = (Z - 1j * n[type_[k]] * tmp[k]) / (1 - 1j * tmp[k] * Z / n[type_[k]])
        # Specific to air.
        r[m] = (1 - Z) / (1 + Z)
    # c=1-np.mean(abs(r)**2)
    c = 1 - np.real(np.sum(r * np.conj(r)) / np.size(lam))
    return float(c)


def cascade(T: np.ndarray, U: np.ndarray) -> np.ndarray:
    n = int(T.shape[1] / 2)
    J = np.linalg.inv(np.eye(n) - np.matmul(U[0:n, 0:n], T[n : 2 * n, n : 2 * n]))
    K = np.linalg.inv(np.eye(n) - np.matmul(T[n : 2 * n, n : 2 * n], U[0:n, 0:n]))
    S = np.block(
        [
            [
                T[0:n, 0:n]
                + np.matmul(
                    np.matmul(np.matmul(T[0:n, n : 2 * n], J), U[0:n, 0:n]),
                    T[n : 2 * n, 0:n],
                ),
                np.matmul(np.matmul(T[0:n, n : 2 * n], J), U[0:n, n : 2 * n]),
            ],
            [
                np.matmul(np.matmul(U[n : 2 * n, 0:n], K), T[n : 2 * n, 0:n]),
                U[n : 2 * n, n : 2 * n]
                + np.matmul(
                    np.matmul(np.matmul(U[n : 2 * n, 0:n], K), T[n : 2 * n, n : 2 * n]),
                    U[0:n, n : 2 * n],
                ),
            ],
        ]
    )
    return S  # type: ignore


def c_bas(A: np.ndarray, V: np.ndarray, h: float) -> np.ndarray:
    n = int(A.shape[1] / 2)
    D = np.diag(np.exp(1j * V * h))
    S = np.block(
        [
            [A[0:n, 0:n], np.matmul(A[0:n, n : 2 * n], D)],
            [
                np.matmul(D, A[n : 2 * n, 0:n]),
                np.matmul(np.matmul(D, A[n : 2 * n, n : 2 * n]), D),
            ],
        ]
    )
    return S  # type: ignore


def marche(a: float, b: float, p: float, n: int, x: float) -> np.ndarray:
    l = np.zeros(n, dtype=np.complex_)  # noqa
    m = np.zeros(n, dtype=np.complex_)
    tmp = (
        1
        / (2 * np.pi * np.arange(1, n))
        * (np.exp(-2 * 1j * np.pi * p * np.arange(1, n)) - 1)
        * np.exp(-2 * 1j * np.pi * np.arange(1, n) * x)
    )
    l[1:n] = 1j * (a - b) * tmp
    l[0] = p * a + (1 - p) * b
    m[0] = l[0]
    m[1:n] = 1j * (b - a) * np.conj(tmp)
    T = toeplitz(l, m)
    return T  # type: ignore


def creneau(
    k0: float, a0: float, pol: float, e1: float, e2: float, a: float, n: int, x0: float
) -> tp.Tuple[np.ndarray, np.ndarray]:
    nmod = int(n / 2)
    alpha = np.diag(a0 + 2 * np.pi * np.arange(-nmod, nmod + 1))
    if pol == 0:
        M = alpha * alpha - k0 * k0 * marche(e1, e2, a, n, x0)
        L, E = np.linalg.eig(M)
        L = np.sqrt(-L + 0j)
        L = (1 - 2 * (np.imag(L) < -1e-15)) * L
        P = np.block([[E], [np.matmul(E, np.diag(L))]])
    else:
        U = marche(1 / e1, 1 / e2, a, n, x0)
        T = np.linalg.inv(U)
        M = (
            np.matmul(
                np.matmul(np.matmul(T, alpha), np.linalg.inv(marche(e1, e2, a, n, x0))),
                alpha,
            )
            - k0 * k0 * T
        )
        L, E = np.linalg.eig(M)
        L = np.sqrt(-L + 0j)
        L = (1 - 2 * (np.imag(L) < -1e-15)) * L
        P = np.block([[E], [np.matmul(np.matmul(U, E), np.diag(L))]])
    return P, L


def homogene(k0: float, a0: float, pol: float, epsilon: float, n: int) -> tp.Tuple[np.ndarray, np.ndarray]:
    nmod = int(n / 2)
    valp = np.sqrt(epsilon * k0 * k0 - (a0 + 2 * np.pi * np.arange(-nmod, nmod + 1)) ** 2 + 0j)
    valp = valp * (1 - 2 * (valp < 0))
    P = np.block([[np.eye(n)], [np.diag(valp * (pol / epsilon + (1 - pol)))]])
    return P, valp


def interface(P: np.ndarray, Q: np.ndarray) -> np.ndarray:
    n = int(P.shape[1])
    S = np.matmul(
        np.linalg.inv(np.block([[P[0:n, 0:n], -Q[0:n, 0:n]], [P[n : 2 * n, 0:n], Q[n : 2 * n, 0:n]]])),
        np.block([[-P[0:n, 0:n], Q[0:n, 0:n]], [P[n : 2 * n, 0:n], Q[n : 2 * n, 0:n]]]),
    )
    return S  # type: ignore


def morpho(X: np.ndarray) -> float:
    lam = 449.5897
    pol = 1.0
    d = 600.521475
    nmod = 25
    # nmod=1
    e2 = 2.4336
    n = 2 * nmod + 1
    n_motifs = int(X.size / 4)
    X = X / d
    h = X[0:n_motifs]
    x0 = X[n_motifs : 2 * n_motifs]
    a = X[2 * n_motifs : 3 * n_motifs]
    spacers = X[3 * n_motifs : 4 * n_motifs]
    l = lam / d  # noqa
    k0 = 2 * np.pi / l
    P, V = homogene(k0, 0, pol, 1, n)
    S = np.block([[np.zeros([n, n]), np.eye(n, dtype=np.complex_)], [np.eye(n), np.zeros([n, n])]])
    for j in range(0, n_motifs):
        Pc, Vc = creneau(k0, 0, pol, e2, 1, a[j], n, x0[j])
        S = cascade(S, interface(P, Pc))
        S = c_bas(S, Vc, h[j])
        S = cascade(S, interface(Pc, P))
        S = c_bas(S, V, spacers[j])
    Pc, Vc = homogene(k0, 0, pol, e2, n)
    S = cascade(S, interface(P, Pc))
    R = np.zeros(3, dtype=np.float_)
    for j in range(-1, 2):
        R[j] = abs(S[j + nmod, nmod]) ** 2 * np.real(V[j + nmod]) / k0
    cost: float = 1 - (R[-1] + R[1]) / 2 + R[0] / 2

    lams = (np.array([400, 500, 600, 700, 800]) + 0.24587) / d
    bar = 0
    for lo in lams:
        k0 = 2 * np.pi / lo
        P, V = homogene(k0, 0, pol, 1, n)
        S = np.block(
            [
                [np.zeros([n, n], dtype=np.complex_), np.eye(n)],
                [np.eye(n), np.zeros([n, n])],
            ]
        )
        for j in range(0, n_motifs):
            Pc, Vc = creneau(k0, 0, pol, e2, 1, a[j], n, x0[j])
            S = cascade(S, interface(P, Pc))
            S = c_bas(S, Vc, h[j])
            S = cascade(S, interface(Pc, P))
            S = c_bas(S, V, spacers[j])
        Pc, Vc = homogene(k0, 0, pol, e2, n)
        S = cascade(S, interface(P, Pc))
        bar += abs(S[nmod, nmod]) ** 2 * np.real(V[nmod]) / k0
    cost += bar / lams.size
    return cost


i = complex(0, 1)


def epscSi(lam: np.ndarray) -> np.ndarray:
    a = np.arange(250, 1500, 5)
    e = np.load(
        Path(__file__).with_name("epsilon_epscSi.npy")
    )  # saved with np.save(filename, e) and dumped in this folder
    y = np.argmin(np.sign(lam - a)) - 1
    epsilon = (e[y + 1] - e[y]) / (a[y + 1] - a[y]) * (lam - a[y]) + e[y]
    return epsilon  # type: ignore


def cascade2(A: np.ndarray, B: np.ndarray) -> np.ndarray:
    """
    This function takes two 2x2 matrices A and B, that are assumed to be scattering matrices
    and combines them assuming A is the "upper" one, and B the "lower" one, physically.
    The result is a 2x2 scattering matrix.
    """
    t = 1 / (1 - B[0, 0] * A[1, 1])
    S = np.zeros((2, 2), dtype=complex)
    S[0, 0] = A[0, 0] + A[0, 1] * B[0, 0] * A[1, 0] * t
    S[0, 1] = A[0, 1] * B[0, 1] * t
    S[1, 0] = B[1, 0] * A[1, 0] * t
    S[1, 1] = B[1, 1] + A[1, 1] * B[0, 1] * B[1, 0] * t
    return S


def solar(lam: np.ndarray) -> np.ndarray:
    # saved with np.save(filename, e) and dumped in this folder
    a = np.load(Path(__file__).with_name("wavelength_solar.npy"))
    e = np.load(Path(__file__).with_name("epsilon_solar.npy"))
    jsc = np.interp(lam, a, e)
    return jsc  # type: ignore


def absorption(
    lam: float,
    epsilon: np.ndarray,
    mu: np.ndarray,
    type_: np.ndarray,
    hauteur: np.ndarray,
    pol: int,
    theta: float,
) -> np.ndarray:
    f = mu if not pol else epsilon
    k0 = 2 * np.pi / lam
    g = type_.size
    alpha = np.sqrt(epsilon[type_[0]] * mu[type_[0]]) * k0 * np.sin(theta)
    gamma = np.sqrt(epsilon[type_] * mu[type_] * k0**2 - np.ones(g) * alpha**2)
    if np.real(epsilon[type_[0]]) < 0 and np.real(mu[type_[0]]) < 0:
        gamma[0] = -gamma[0]
    if g > 2:
        gamma[1 : g - 2] = gamma[1 : g - 2] * (1 - 2 * (np.imag(gamma[1 : g - 2]) < 0))
    if (
        np.real(epsilon[type_[g - 1]]) < 0
        and np.real(mu[type_[g - 1]]) < 0
        and np.real(np.sqrt(epsilon[type_[g - 1]] * mu[type_[g - 1]] * k0**2 - alpha**2)) != 0
    ):
        gamma[g - 1] = -np.sqrt(epsilon[type_[g - 1]] * mu[type_[g - 1]] * k0**2 - alpha**2)
    else:
        gamma[g - 1] = np.sqrt(epsilon[type_[g - 1]] * mu[type_[g - 1]] * k0**2 - alpha**2)
    T = np.zeros(((2 * g, 2, 2)), dtype=complex)
    T[0] = [[0, 1], [1, 0]]
    for k2 in range(g - 1):
        t = np.exp(i * gamma[k2] * hauteur[k2])
        T[2 * k2 + 1] = [[0, t], [t, 0]]
        # Interface scattering matrix
        b1 = gamma[k2] / f[type_[k2]]
        b2 = gamma[k2 + 1] / f[type_[k2 + 1]]
        T[2 * k2 + 2] = [
            [(b1 - b2) / (b1 + b2), 2 * b2 / (b1 + b2)],
            [2 * b1 / (b1 + b2), (b2 - b1) / (b1 + b2)],
        ]
        t = np.exp(i * gamma[g - 1] * hauteur[g - 1])
        T[2 * g - 1] = [[0, t], [t, 0]]
    H = np.zeros(((2 * g - 1, 2, 2)), dtype=complex)
    A = np.zeros(((2 * g - 1, 2, 2)), dtype=complex)
    H[0] = T[2 * g - 1]
    A[0] = T[0]
    for j in range(2 * g - 2):
        A[j + 1] = cascade2(A[j], T[j + 1])
        H[j + 1] = cascade2(T[2 * g - 2 - j], H[j])
    # r = A[len(A) - 1][0, 0]  # TODO: unused
    t = A[len(A) - 1][1, 0]
    I = np.zeros(((2 * g, 2, 2)), dtype=complex)  # noqa
    for j in range(len(T) - 1):
        I[j][0, 0] = A[j][1, 0] / (1 - A[j][1, 1] * H[len(T) - 2 - j][0, 0])
        I[j][0, 1] = A[j][1, 1] * H[len(T) - 2 - j][0, 1] / (1 - A[j][1, 1] * H[len(T) - 2 - j][0, 0])
        I[j][1, 0] = A[j][1, 0] * H[len(T) - 2 - j][0, 0] / (1 - A[j][1, 1] * H[len(T) - 2 - j][0, 0])
        I[j][1, 1] = H[len(T) - 2 - j][0, 1] / (1 - A[j][1, 1] * H[len(T) - 2 - j][0, 0])
    I[2 * g - 1][0, 0] = I[2 * g - 2][0, 0] * np.exp(i * gamma[g - 1] * hauteur[g - 1])
    I[2 * g - 1][0, 1] = I[2 * g - 2][0, 1] * np.exp(i * gamma[g - 1] * hauteur[g - 1])
    I[2 * g - 1][1, 0] = 0
    I[2 * g - 1][1, 1] = 0
    w = 0
    poynting = np.zeros(2 * g, dtype=complex)
    if pol == 0:  # TE
        for j in range(2 * g):
            poynting[j] = np.real(
                (I[j][0, 0] + I[j][1, 0])
                * np.conj((I[j][0, 0] - I[j][1, 0]) * gamma[w] / mu[type_[w]])
                * mu[type_[0]]
                / (gamma[0])
            )
            w = w + 1 - np.mod(j + 1, 2)
    else:  # TM
        for j in range(2 * g):
            poynting[j] = np.real(
                (I[j][0, 0] - I[j][1, 0])
                * np.conj((I[j][0, 0] + I[j][1, 0]) * gamma[w] / epsilon[type_[w]])
                * epsilon[type_[0]]
                / (gamma[0])
            )
            w = w + 1 - np.mod(j + 1, 2)
    tmp = abs(-np.diff(poynting))
    absorb = tmp[np.arange(0, 2 * g, 2)]
    return absorb  # type: ignore


def cf_photosic_reference(X: np.ndarray) -> float:
    """vector X is only the thicknesses of each layers, because the materials (so the epislon)
    are imposed by the function. This is similar in the chirped function.
    """
    lam_min = 375
    lam_max = 750
    n_lam = 100
    theta = 0 * np.pi / 180
    vlam = np.linspace(lam_min, lam_max, n_lam)
    scc = np.zeros(n_lam)
    Ab = np.zeros(n_lam)
    for k in range(n_lam):
        lam = vlam[k]
        epsilon = np.array([1, 2, 3, epscSi(lam)], dtype=complex)
        mu = np.ones(epsilon.size)
        type_ = np.append(0, np.append(np.tile(np.array([1, 2]), int(X.size / 2)), 3))
        hauteur = np.append(0, np.append(X, 30000))
        pol = 0
        absorb = absorption(lam, epsilon, mu, type_, hauteur, pol, theta)
        scc[k] = solar(lam)
        Ab[k] = absorb[len(absorb) - 1]
    max_scc = np.trapz(scc, vlam)
    j_sc = np.trapz(scc * Ab, vlam)
    CE = j_sc / max_scc
    cost = 1 - CE
    return cost  # type: ignore


def cf_photosic_realistic(eps_and_d: np.ndarray) -> float:
    """eps_and_d is a vector composed in a first part with the epsilon values
    (the material used in each one of the layers), and in a second part with the
    thicknesses of each one of the layers, like in Bragg.
    Any number of layers can work. Basically I used between 4 and 50 layers,
    and the best results are generally obtained when the structure has between 10 and 20 layers.
    The epsilon values are generally comprised between 1.00 and 9.00.
    """
    dimension = int(eps_and_d.size / 2)
    eps = eps_and_d[0:dimension]
    d = eps_and_d[dimension : dimension * 2]
    epsd = np.array([eps, d])
    lam_min = 375
    lam_max = 750
    n_lam = 100
    theta = 0 * 180 / np.pi
    vlam = np.linspace(lam_min, lam_max, n_lam)
    scc = np.zeros(n_lam)
    Ab = np.zeros(n_lam)
    # spectrum=np.zeros(n_lam)
    for k in range(n_lam):
        # absorb=absorption(epsd,theta*pi/180,lam[k])
        lam = vlam[k]
        epsilon = np.append(1, np.append(epsd[0], epscSi(lam)))
        mu = np.ones(epsilon.size)
        type_ = np.arange(0, epsd[0].size + 2)
        hauteur = np.append(0, np.append(epsd[1], 30000))
        pol = 0
        absorb = absorption(lam, epsilon, mu, type_, hauteur, pol, theta)
        scc[k] = solar(lam)
        Ab[k] = absorb[len(absorb) - 1]
    max_scc = np.trapz(scc, vlam)
    j_sc = np.trapz(scc * Ab, vlam)
    CE = j_sc / max_scc
    cost = 1 - CE
    return cost  # type: ignore



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""Definitions of some convenient types.
If you know better practices, feel free to submit it ;)
"""
# pylint: disable=unused-import
# structures
from typing import Any as Any
from typing import Generic as Generic
from typing import Type as Type
from typing import TypeVar as TypeVar
from typing import Optional as Optional
from typing import Union as Union

# containers
from typing import Dict as Dict
from typing import Tuple as Tuple
from typing import List as List
from typing import Set as Set
from typing import Deque as Deque
from typing import Sequence as Sequence
from typing import NamedTuple as NamedTuple
from typing import MutableMapping as MutableMapping

# iterables
from typing import Iterator as Iterator
from typing import Iterable as Iterable
from typing import Generator as Generator
from typing import KeysView as KeysView
from typing import ValuesView as ValuesView
from typing import ItemsView as ItemsView

# others
from typing import Callable as Callable
from typing import Hashable as Hashable
from typing import Match as Match
from typing import cast as cast
from pathlib import Path as Path

try:
    from typing import Protocol
except:
    from typing_extensions import Protocol  # type: ignore

#
import numpy as _np


ArgsKwargs = Tuple[Tuple[Any, ...], Dict[str, Any]]
ArrayLike = Union[Tuple[float, ...], List[float], _np.ndarray]  # most common
PathLike = Union[str, Path]
FloatLoss = float
Loss = Union[float, ArrayLike]
BoundValue = Optional[Union[float, int, _np.int_, _np.float_, ArrayLike]]


# %% Protocol definitions for executor typing

X = TypeVar("X", covariant=True)


class JobLike(Protocol[X]):
    # pylint: disable=pointless-statement

    def done(self) -> bool:
        ...

    def result(self) -> X:
        ...


class ExecutorLike(Protocol):
    # pylint: disable=pointless-statement, unused-argument

    def submit(self, fn: Callable[..., X], *args: Any, **kwargs: Any) -> JobLike[X]:
        ...



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from . import plotting  # pylint: disable=wrong-import-position, wrong-import-order
from unittest.mock import patch
from pathlib import Path
import typing as tp
import pytest
import numpy as np
import pandas as pd
import matplotlib
from nevergrad.common import testing
from . import utils

matplotlib.use("Agg")


def test_get_winners_df() -> None:
    dfdata = [["alg0", 46424, 0.4], ["alg1", 4324546, 0.1], ["alg1", 424546, 0.5], ["alg2", 1424546, 0.3]]
    df = pd.DataFrame(columns=["optimizer_name", "blublu", "loss"], data=dfdata)
    all_optimizers = [f"alg{k}" for k in range(4)]
    # blublu column is useless, and losses are meaned for each algorithm
    winners = plotting._make_winners_df(df, all_optimizers)
    data = 0.5 * np.identity(4)  # self playing is a tie
    # unspecified algo
    data[(3, 3)] = 0  # type: ignore
    # alg1 and alg2 win over alg0  # type: ignore
    data[tuple(zip(*[(1, 0), (2, 0)]))] = 1  # type: ignore
    # alg1 and alg2 are a tie (mean loss .3)
    data[tuple(zip(*[(1, 2), (2, 1)]))] = 0.5  # type: ignore
    expected = pd.DataFrame(index=all_optimizers, columns=all_optimizers, data=data)
    winners.assert_equivalent(expected)


def test_make_sorted_winrates() -> None:
    algos = [f"alg{k}" for k in range(4)]
    data = [
        [0, 0, 0, 0],  # unplayed
        [0, 2, 4, 4],  # all time winner (4 games
        [0, 0, 2, 1],  # last
        [0, 0, 3, 2],
    ]
    victories = pd.DataFrame(index=algos, columns=algos, data=data)
    winrates = plotting._make_sorted_winrates_df(victories)
    expected_data = [[0.5, 1, 1, -1.0], [0, 0.5, 0.75, -1], [0, 0.25, 0.5, -1], [0, 0, 0, -1]]
    winrates = winrates.fillna(-1)
    salgos = [f"alg{k}" for k in [1, 3, 2, 0]]
    expected = pd.DataFrame(index=salgos, columns=salgos, data=expected_data)
    assert winrates.equals(expected), f"Expected:\n{expected}\nbut got:\n{winrates}"


def test_create_plots_from_csv_mocked() -> None:
    df = pd.read_csv(Path(__file__).parent / "sphere_perf_example.csv")
    with patch("nevergrad.benchmark.plotting.XpPlotter"):
        with patch("nevergrad.benchmark.plotting.FightPlotter") as fplt:
            plotting.create_plots(df, "", max_combsize=1)
            assert (
                fplt.call_count == 12
            ), "Should be called for all, 2 noise levels and 3 budgets, and, for each, a ``pure'' context."


def test_fight_plotter() -> None:
    df = utils.Selector.read_csv(Path(__file__).parent / "sphere_perf_example.csv").select(
        optimizer_name=["OnePlusOneOptimizer", "HaltonSearch", "Powell"]
    )
    winrates = plotting.FightPlotter.winrates_from_selection(df, ["noise_level", "budget"])
    # check data
    np.testing.assert_array_equal(
        winrates.index,
        [
            "Powell (1/3:75.0% +- 4.4)",
            "OnePlusOneOptimizer (2/3:58.3% +- 5.0)",
            "Halton (3/3:16.7% +- 3.8)",
        ],
    )
    np.testing.assert_array_equal(
        winrates.columns, ["Powell (6/6)", "OnePlusOneOptimizer (6/6)", "HaltonSearch (6/6)"]
    )
    np.testing.assert_almost_equal(winrates, [[0.5, 0.75, 1], [0.25, 0.5, 1], [0, 0, 0.5]])
    # plot
    plotter = plotting.FightPlotter(winrates)
    with patch("matplotlib.pyplot.Figure.savefig"):
        plotter.save("should_not_exist.png")


def test_xp_plotter() -> None:
    opt = "OnePlusOneOptimizer"
    df = utils.Selector.read_csv(Path(__file__).parent / "sphere_perf_example.csv").select(
        optimizer_name=[opt]
    )
    unused_data = plotting.XpPlotter.make_data(df, normalized_loss=True)
    data = plotting.XpPlotter.make_data(df)
    # check data
    testing.assert_set_equal(data.keys(), {opt})
    testing.assert_set_equal(data[opt].keys(), {"budget", "loss", "loss_std", "num_eval"})
    np.testing.assert_almost_equal(data[opt]["budget"], [200, 400, 800])
    np.testing.assert_almost_equal(data[opt]["loss"], [0.4811605, 0.3920045, 0.14778369])
    np.testing.assert_almost_equal(data[opt]["loss_std"], [0.83034832, 0.73255529, 0.18551625])
    # plot
    with patch("matplotlib.pyplot.Figure.tight_layout"):  # avoid warning message
        plotter = plotting.XpPlotter(data, title="Title")
    with patch("matplotlib.pyplot.Figure.savefig"):
        plotter.save("should_not_exist.png")


def test_remove_errors() -> None:
    data = [
        ["alg0", 0, 10, np.nan],
        ["alg2", np.nan, 30, "ValueError"],
        ["alg1", 0, 20, "SomeHandledError"],
        ["alg3", np.nan, 40, "BlubluError"],
    ]
    df = pd.DataFrame(columns=["optimizer_name", "loss", "dimension", "error"], data=data)
    with pytest.warns(UserWarning) as w:
        output = plotting.remove_errors(df)
    assert len(w) == 3
    expected = pd.DataFrame(
        columns=["optimizer_name", "loss", "dimension"], data=[["alg0", 0, 10], ["alg1", 0, 20]]
    )
    np.testing.assert_array_equal(output.columns, expected.columns)
    np.testing.assert_array_equal(output.index, expected.index)
    np.testing.assert_array_equal(output, expected)
    assert isinstance(output, plotting.utils.Selector)


def test_remove_nan_value() -> None:
    data = [["alg0", 0, 10, np.nan], ["alg2", np.nan, 30, np.nan]]
    df = pd.DataFrame(columns=["optimizer_name", "loss", "dimension", "error"], data=data)
    with pytest.warns(UserWarning) as w:
        output = plotting.remove_errors(df)
    assert len(w) == 1
    expected = pd.DataFrame(columns=["optimizer_name", "loss", "dimension"], data=[["alg0", 0, 10]])
    np.testing.assert_array_equal(output, expected)


def test_make_style_generator() -> None:
    num = 364
    gen = plotting._make_style_generator()
    output = [next(gen) for _ in range(num)]
    np.testing.assert_equal(output[:5], ["-ob", "--vg", ":^r", "-.<c", "->m"])
    # the following is only out of curiosity
    np.testing.assert_equal(len(set(output)), num)  # no repetition
    repeating = next(gen)
    np.testing.assert_equal(repeating, output[0])


def test_name_style() -> None:
    nstyle = plotting.NameStyle()
    np.testing.assert_equal(nstyle["blublu"], "-ob")
    np.testing.assert_equal(nstyle["plop"], "--vg")
    np.testing.assert_equal(nstyle["blublu"], "-ob")


def test_split_long_title() -> None:
    title = "abcd,efgh"
    np.testing.assert_equal(plotting.split_long_title(title), title)
    title = ",".join(["a" * 25, "b" * 25, "c" * 25, "d" * 15])
    np.testing.assert_equal(plotting.split_long_title(title), title[:52] + "\n" + title[52:])
    title = "a" * 70
    np.testing.assert_equal(plotting.split_long_title(title), title)


@testing.parametrized(
    nothing=([1, 2, 10.0], [1, 2, 10.0]),
    identic=([1, 1, 10.0, 10.0], [0.5, 1.5, 9.5, 10.5]),
)
def test_compute_best_placements(positions: tp.List[float], expected: tp.List[float]) -> None:
    new_positions = plotting.compute_best_placements(positions, min_diff=1.0)
    np.testing.assert_array_equal(new_positions, expected)


def test_merge_parametrization_and_optimizer() -> None:
    df = pd.DataFrame(
        columns=["optimizer_name", "parametrization", "val"],
        data=[["o1", "p1", 1], ["o1", "p2", 2], ["o2", np.nan, 3]],
    )
    out = plotting.merge_optimizer_name_pattern(utils.Selector(df), "{optimizer_name},{parametrization}")
    assert isinstance(out, utils.Selector)
    assert out["optimizer_name"].tolist() == ["o1,p1", "o1,p2", "o2"]
    assert out["val"].tolist() == [1, 2, 3]
    # nothing to do
    out = plotting.merge_optimizer_name_pattern(utils.Selector(df), "")
    assert out["optimizer_name"].tolist() == ["o1", "o1", "o2"]


def test_normalized_losses() -> None:
    data = [
        ["alg0", 0, "sphere", 3],
        ["alg0", -2, "sphere", 3],
        ["alg2", 2, "sphere", 3],
        ["alg3", 12, "sphere", 12],
        ["alg2", 5, "sphere", 12],
        ["alg4", 24, "ellipsoid", 3],
        ["alg5", float("inf"), "sphere", 3],
    ]
    df = pd.DataFrame(columns=["optimizer_name", "loss", "func", "dimension"], data=data)
    ndf = plotting.normalized_losses(df, ["func", "dimension"])
    np.testing.assert_array_equal(ndf.loss, [0.5, 0, 1, 1, 0, 1, float("inf")])


if __name__ == "__main__":
    # simple example which can be run with:
    # python -m nevergrad.benchmark.test_plotting
    df_test = pd.read_csv(Path(__file__).parent / "sphere_perf_example.csv")
    plotting.create_plots(df_test, output_folder="", max_combsize=0)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import site
import glob
import ctypes
import time
import inspect
import warnings
import itertools
import contextlib
import collections
import typing as tp
import numpy as np
from nevergrad.common import sphere


def quasi_randomize(x: tp.Iterable[tp.Any], method: str = "none") -> tp.Any:
    return sphere.quasi_randomize(x, method)


def pytorch_import_fix() -> None:
    """Hackfix needed before pytorch import ("dlopen: cannot load any more object with static TLS")
    See issue #305
    """
    try:
        for packages in site.getsitepackages():
            for lib in glob.glob(f"{packages}/torch/lib/libgomp*.so*"):
                ctypes.cdll.LoadLibrary(lib)
    except Exception:  # pylint: disable=broad-except
        pass


def pairwise(iterable: tp.Iterable[tp.Any]) -> tp.Iterator[tp.Tuple[tp.Any, tp.Any]]:
    """Returns an iterator over sliding pairs of the input iterator
    s -> (s0,s1), (s1,s2), (s2, s3), ...

    Note
    ----
    Nothing will be returned if length of iterator is strictly less
    than 2.
    """  # From itertools documentation
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def grouper(iterable: tp.Iterable[tp.Any], n: int, fillvalue: tp.Any = None) -> tp.Iterator[tp.List[tp.Any]]:
    """Collect data into fixed-length chunks or blocks
    Copied from itertools recipe documentation
    Example: grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    """
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)


def roundrobin(*iterables: tp.Iterable[tp.Any]) -> tp.Iterator[tp.Any]:
    """roundrobin('ABC', 'D', 'EF') --> A D E B F C"""
    # Recipe credited to George Sakkis
    num_active = len(iterables)
    nexts = itertools.cycle(iter(it).__next__ for it in iterables)
    while num_active:
        try:
            for next_ in nexts:
                yield next_()
        except StopIteration:
            # Remove the iterator we just exhausted from the cycle.
            num_active -= 1
            nexts = itertools.cycle(itertools.islice(nexts, num_active))


class Sleeper:
    """Simple object for managing the waiting time of a job

    Parameters
    ----------
    min_sleep: float
        minimum sleep time
    max_sleep: float
        maximum sleep time
    averaging_size: int
        size for averaging the registered durations
    """

    def __init__(self, min_sleep: float = 1e-7, max_sleep: float = 1.0, averaging_size: int = 10) -> None:
        self._min = min_sleep
        self._max = max_sleep
        self._start: tp.Optional[float] = None
        self._queue: tp.Deque[float] = collections.deque(maxlen=averaging_size)
        self._num_waits = 10  # expect to waste around 10% of time

    def start_timer(self) -> None:
        if self._start is not None:
            warnings.warn("Ignoring since timer was already started.")
            return
        self._start = time.time()

    def stop_timer(self) -> None:
        if self._start is None:
            warnings.warn("Ignoring since timer was stopped before starting.")
            return
        self._queue.append(time.time() - self._start)
        self._start = None

    def _get_advised_sleep_duration(self) -> float:
        if not self._queue:
            if self._start is None:
                return self._min
            value = time.time() - self._start
        else:
            value = float(np.mean(self._queue))  # type: ignore
        return float(np.clip(value / self._num_waits, self._min, self._max))

    def sleep(self) -> None:
        time.sleep(self._get_advised_sleep_duration())


X = tp.TypeVar("X", bound=tp.Hashable)


class OrderedSet(tp.MutableSet[X]):
    """Set of elements retaining the insertion order
    All new elements are appended to the end of the set.
    """

    def __init__(self, keys: tp.Optional[tp.Iterable[X]] = None) -> None:
        self._data: "collections.OrderedDict[X, int]" = collections.OrderedDict()
        self._global_index = 0  # keep track of insertion global index if need be
        if keys is not None:
            for key in keys:
                self.add(key)

    def add(self, key: X) -> None:
        self._data[key] = self._data.pop(key, self._global_index)
        self._global_index += 1

    def popright(self) -> X:
        key = next(reversed(self._data))
        self.discard(key)
        return key

    def discard(self, key: X) -> None:
        del self._data[key]

    def __contains__(self, key: tp.Any) -> bool:
        return key in self._data

    def __iter__(self) -> tp.Iterator[X]:
        return iter(self._data)

    def __len__(self) -> int:
        return len(self._data)


def different_from_defaults(
    *,
    instance: tp.Any,
    instance_dict: tp.Optional[tp.Dict[str, tp.Any]] = None,
    check_mismatches: bool = False,
) -> tp.Dict[str, tp.Any]:
    """Checks which attributes are different from defaults arguments

    Parameters
    ----------
    instance: object
        the object to change
    instance_dict: dict
        the dict corresponding to the instance, if not provided it's self.__dict__
    check_mismatches: bool
        checks that the attributes match the parameters

    Note
    ----
    This is convenient for short repr of data structures
    """
    defaults = {
        x: y.default
        for x, y in inspect.signature(instance.__class__.__init__).parameters.items()
        if x not in ["self", "__class__"]
    }
    if instance_dict is None:
        instance_dict = instance.__dict__
    if check_mismatches:
        add = set(defaults.keys()) - set(instance_dict.keys())
        miss = set(instance_dict.keys()) - set(defaults.keys())
        if add or miss:  # this is to help during development
            raise RuntimeError(
                f"Mismatch between attributes and arguments of {instance.__class__}:\n"
                f"- additional: {add}\n- missing: {miss}"
            )
    else:
        defaults = {x: y for x, y in defaults.items() if x in instance.__dict__}
    # only print non defaults
    return {
        x: instance_dict[x] for x, y in defaults.items() if y != instance_dict[x] and not x.startswith("_")
    }


@contextlib.contextmanager
def set_env(**environ: tp.Any) -> tp.Generator[None, None, None]:
    """Temporarily changes environment variables."""
    old_environ = {x: os.environ.get(x, None) for x in environ}
    for x in environ:
        if x != x.upper():
            raise ValueError(f"Only capitalized environment variable are allowed, but got {x!r}")
    os.environ.update({x: str(y) for x, y in environ.items()})
    try:
        yield
    finally:
        for k, val in old_environ.items():
            os.environ.pop(k)
            if val is not None:
                os.environ[k] = val


def flatten(obj: tp.Any) -> tp.Any:
    """Flatten a dict/list structure

    Example
    -------

    >>> flatten(["a", {"truc": [4, 5]}])
    >>> {"0": "a", "1.truc.0": 4, "1.truc.1": 5}
    """
    output: tp.Any = {}
    if isinstance(obj, (tuple, list)):
        iterator = enumerate(obj)
    elif isinstance(obj, dict):
        iterator = obj.items()  # type: ignore
    else:
        return obj
    for k, val in iterator:
        content = flatten(val)
        if isinstance(content, dict):
            output.update({f"{k}.{x}": y for x, y in content.items()})
        else:
            output[str(k)] = val
    return output



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import heapq
import typing as tp
from collections import deque
from nevergrad.functions import ExperimentFunction


class MockedTimedJob:
    """Job returned by the MockedTimedExecutor, with the usual
    "done()" and "result()" methods.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        func: tp.Callable[..., tp.Any],
        args: tp.Tuple[tp.Any, ...],
        kwargs: tp.Dict[str, tp.Any],
        executor: "MockedTimedExecutor",
    ) -> None:
        self._executor = executor
        self._time = executor.time  # time at instantiation
        # function
        self._func = func
        self._args = args
        self._kwargs = kwargs
        # job specific
        self._output: tp.Any = None
        self._delay: tp.Optional[float] = None  # float when properly job is computed
        self._done = False
        self._is_read = False  # for the record

    @property
    def release_time(self) -> float:
        self.process()
        assert self._delay is not None
        return self._delay + self._time

    def done(self) -> bool:
        return self._executor.check_is_done(self)

    def process(self) -> None:
        if self._delay is None:
            self._output = self._func(*self._args, **self._kwargs)
            # compute the delay and add to queue
            self._delay = 1.0
            if isinstance(self._func, ExperimentFunction):
                self._delay = max(0, self._func.compute_pseudotime((self._args, self._kwargs), self._output))

    def result(self) -> tp.Any:
        """Return the result if "done()" is true, and raises
        a RuntimeError otherwise.
        """
        self.process()
        if not self.done():
            raise RuntimeError("Asking result which is not ready")
        self._is_read = True
        self._executor.notify_read(self)
        return self._output

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<value={self._output}, done={self._done}, read={self._is_read}>)"


class OrderedJobs(tp.NamedTuple):
    """Handle for sorting jobs by release_time (or submission order in case of tie)"""

    release_time: float
    order: int
    job: MockedTimedJob


class MockedTimedExecutor:
    """Executor that mocks a steady state, by only providing 1 job at a time which is done() while
    not having been "read" (i.e. "result()" method was not executed).
    This ensures we control the order of update of the optimizer for benchmarking.

    Additionally, "delays" can be provided by the function so that jobs are not "done()" by order of
    submission. To this end, callables must implement a "computation_time" method.
    """

    def __init__(self, batch_mode: bool = False) -> None:
        self.batch_mode = batch_mode
        self._to_be_processed: tp.Deque[MockedTimedJob] = deque()
        self._steady_priority_queue: tp.List[OrderedJobs] = []
        self._order = 0
        self._time = 0.0

    @property
    def time(self) -> float:
        return self._time

    def submit(self, fn: tp.Callable[..., tp.Any], *args: tp.Any, **kwargs: tp.Any) -> MockedTimedJob:
        job = MockedTimedJob(fn, args, kwargs, self)
        self._to_be_processed.append(job)  # save for later processing
        return job

    def _process_submissions(self) -> None:
        """Process all submissions which have not been processed yet."""
        while self._to_be_processed:
            job = self._to_be_processed[0]
            job.process()  # trigger computation
            if not self.batch_mode:
                heapq.heappush(self._steady_priority_queue, OrderedJobs(job.release_time, self._order, job))
            self._to_be_processed.popleft()  # remove right after it is added to the heap queue
            self._order += 1

    def check_is_done(self, job: MockedTimedJob) -> bool:
        """Called whenever "done" method is called on a job."""
        self._process_submissions()  # make sure everything is up to date
        if self.batch_mode or job._is_read:
            return True
        else:
            return job is self._steady_priority_queue[0].job

    def notify_read(self, job: MockedTimedJob) -> None:
        """Called whenever a result is read, so as to activate the next result in line
        in case of steady mode, and to update executor time.
        """
        self._process_submissions()  # make sure everything is up to date
        if not self.batch_mode:
            expected = self._steady_priority_queue[0]
            assert job is expected.job, "Only first job should be read"
            heapq.heappop(self._steady_priority_queue)
            if self._steady_priority_queue:
                self._steady_priority_queue[0].job._done = True
        self._time = max(self._time, job.release_time)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import warnings
import nevergrad.common.typing as tp

# import numpy as np
from nevergrad.parametrization import parameter as p
from nevergrad.optimization.utils import UidQueue
from . import base
from .multiobjective import nsga2 as nsga2


class _EvolutionStrategy(base.Optimizer):
    """Experimental evolution-strategy-like algorithm
    The behavior is going to evolve
    """

    def __init__(
        self,
        parametrization: base.IntOrParameter,
        budget: tp.Optional[int] = None,
        num_workers: int = 1,
        *,
        config: tp.Optional["EvolutionStrategy"] = None,
    ) -> None:
        if budget is not None and budget < 60:
            warnings.warn(
                "ES algorithms are inefficient with budget < 60", base.errors.InefficientSettingsWarning
            )
        super().__init__(parametrization, budget=budget, num_workers=num_workers)
        self._population: tp.Dict[str, p.Parameter] = {}
        self._uid_queue = UidQueue()
        self._waiting: tp.List[p.Parameter] = []
        # configuration
        self._config = EvolutionStrategy() if config is None else config
        self._rank_method: tp.Any = None  # TODO better typing (eventually)
        if self._config.ranker == "nsga2":
            self._rank_method = nsga2.rank
        elif self._config.ranker != "simple":
            raise NotImplementedError(f"Unknown ranker {self._config.ranker}")
        self._no_hypervolume = self._config.offsprings is None

    def _internal_ask_candidate(self) -> p.Parameter:
        if self.num_ask < self._config.popsize:
            param = self.parametrization.sample()
            assert param.uid == param.heritage["lineage"]  # this is an assumption used below
            self._uid_queue.asked.add(param.uid)
            self._population[param.uid] = param
            return param
        uid = self._uid_queue.ask()
        param = self._population[uid].spawn_child()
        param.mutate()
        ratio = self._config.recombination_ratio
        if ratio and self._rng.rand() < ratio:
            selected = self._rng.choice(list(self._population))
            param.recombine(self._population[selected])
        return param

    def _internal_tell_candidate(self, candidate: p.Parameter, loss: float) -> None:
        if self._config.offsprings is None:
            uid = candidate.heritage["lineage"]
            self._uid_queue.tell(uid)
            parent_value = float("inf") if uid not in self._population else base._loss(self._population[uid])
            if loss < parent_value:
                self._population[uid] = candidate
        else:
            no_parent = next(iter(candidate.parents_uids), "#no_parent#") not in self._population
            if no_parent and len(self._population) < self._config.popsize:
                self._population[candidate.uid] = candidate
                self._uid_queue.tell(candidate.uid)
            else:
                self._waiting.append(candidate)
            if len(self._waiting) >= self._config.offsprings:
                self._select()

    def _select(self) -> None:
        choices = self._waiting + ([] if self._config.only_offsprings else list(self._population.values()))
        if self._rank_method is not None and self.num_objectives > 1:
            choices_rank = self._rank_method(choices, n_selected=self._config.popsize)
            choices = [x for x in choices if x.uid in choices_rank]
        else:
            choices.sort(key=base._loss)
        self._population = {x.uid: x for x in choices[: self._config.popsize]}
        self._uid_queue.clear()
        self._waiting.clear()
        for uid in self._population:
            self._uid_queue.tell(uid)


class EvolutionStrategy(base.ConfiguredOptimizer):
    """Experimental evolution-strategy-like algorithm
    The API is going to evolve

    Parameters
    ----------
    recombination_ratio: float
        probability of using a recombination (after the mutation) for generating new offsprings
    popsize: int
        population size of the parents (lambda)
    offsprings: int
        number of generated offsprings (mu)
    only_offsprings: bool
        use only offsprings for the new generation if True (True: lambda,mu, False: lambda+mu)
    ranker: str
        ranker for the multiobjective case (defaults to NSGA2)
    """

    # pylint: disable=unused-argument
    def __init__(
        self,
        *,
        recombination_ratio: float = 0,
        popsize: int = 40,
        offsprings: tp.Optional[int] = None,
        only_offsprings: bool = False,
        ranker: str = "nsga2",
    ) -> None:
        super().__init__(_EvolutionStrategy, locals(), as_config=True)
        assert offsprings is None or not only_offsprings or offsprings > popsize
        if only_offsprings:
            assert offsprings is not None, "only_offsprings only work if offsprings is not None (non-DE mode)"
        assert 0 <= recombination_ratio <= 1
        assert ranker in ["simple", "nsga2"]
        self.recombination_ratio = recombination_ratio
        self.popsize = popsize
        self.offsprings = offsprings
        self.only_offsprings = only_offsprings
        self.ranker = ranker


RecES = EvolutionStrategy(recombination_ratio=1, only_offsprings=True, offsprings=60).set_name(
    "RecES", register=True
)
RecMixES = EvolutionStrategy(recombination_ratio=1, only_offsprings=False, offsprings=20).set_name(
    "RecMixES", register=True
)
RecMutDE = EvolutionStrategy(recombination_ratio=1, only_offsprings=False, offsprings=None).set_name(
    "RecMutDE", register=True
)
ES = EvolutionStrategy(recombination_ratio=0, only_offsprings=True, offsprings=60).set_name(
    "ES", register=True
)
MixES = EvolutionStrategy(recombination_ratio=0, only_offsprings=False, offsprings=20).set_name(
    "MixES", register=True
)
MutDE = EvolutionStrategy(recombination_ratio=0, only_offsprings=False, offsprings=None).set_name(
    "MutDE", register=True
)
NonNSGAIIES = EvolutionStrategy(
    recombination_ratio=0, only_offsprings=True, offsprings=60, ranker="simple"
).set_name("NonNSGAIIES", register=True)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from typing import Callable, Optional, List, Iterable, Type, Union
import numpy as np
import pytest
from nevergrad.common import testing
from . import _core


@testing.parametrized(
    sphere=(_core.sphere, 9.043029, None),
    elliptic=(_core.elliptic, 262389.541996, None),
    ackley=(_core.ackley, 4.7298473, None),
    schwefel=(_core.schwefel_1_2, 128.709765, None),
    rosenbrock=(_core.rosenbrock, 1967.57859, None),
    sphere_m=(_core.sphere, 30, [1, 2, 3, 4]),
    elliptic_m=(_core.elliptic, 16090401, [1, 2, 3, 4]),
    ackley_m=(_core.ackley, 2.254444, [0.1, 0.2, 0.3]),
    schwefel_m=(_core.schwefel_1_2, 146, [1, 2, 3, 4]),
    rosenbrock_m=(_core.rosenbrock, 2705, [1, 2, 3, 4]),
)
def test_core_function(
    func: Callable[[np.ndarray], float], expected: float, data: Optional[List[float]]
) -> None:
    if data is None:
        data = [
            0.662,
            -0.217,
            -0.968,
            1.867,
            0.101,
            0.575,
            0.199,
            1.576,
            1.006,
            0.182,
            -0.092,
            0.466,
        ]  # only for non-regression
    value = func(np.array(data))
    np.testing.assert_almost_equal(value, expected, decimal=5)


@testing.parametrized(
    irregularity=(
        _core.irregularity,
        [-10, -2, -1, 0, 0.1, 1, 2, 2.5],
        [-10.426, -2.021, -1, 0, 0.107, 1, 1.988, 2.635],
    ),
    asymmetry_04=(
        _core.Asymmetry(0.2),
        [-10, -2, -1, 0, 0.1, 1, 2, 2.5],
        [-10, -2, -1, 0, 0.092, 1, 2.366, 3.34],
    ),
    asymmetry_02=(_core.Asymmetry(0.4), [-1, 0, 0.1, 1, 2], [-1, 0, 0.086, 1, 2.96]),
    illconditionning_10=(
        _core.Illconditionning(10.0),
        [-2, -1, 0, 0.1, 1, 2],
        [-2.0, -1.259, 0.0, 0.2, 2.512, 6.325],
    ),
    illconditionning_1=(
        _core.Illconditionning(1.0),
        [-2, -1, 0, 0.1, 1, 2],
        [-2, -1, 0, 0.1, 1, 2],
    ),  # no change expected
    translation=(_core.Translation(np.array([1, 2])), [0, 10], [-1, 8]),
    rotation=(_core.Rotation(np.array([[0, 1], [1, 0]])), [0, 10], [10, 0]),
)
def test_transform(
    func: Callable[[np.ndarray], np.ndarray], data: List[float], expected: List[float]
) -> None:
    value = func(np.array(data))
    np.testing.assert_almost_equal(value, expected, decimal=3)


@testing.parametrized(
    simple=(range(8), [3, 3, 2], 0, [[0, 1, 2], [3, 4, 5], [6, 7]]),
    bad_dim=(range(8), [3, 3, 4], 0, ValueError),
    bad_dim_overlap=(range(6), [3, 3, 1], 1, ValueError),
    overlap=(np.arange(6), [3, 3, 2], 1, [[0, 1, 2], [2, 3, 4], [4, 5]]),
    permut=([0, 2, 3, 5, 1, 4], [3, 2, 3], 1, [[0, 2, 3], [3, 5], [5, 1, 4]]),
    wrong_start=(1 + np.arange(6), [3, 3, 2], 1, AssertionError),
)
def test_split(
    permutation: Iterable[float],
    dimensions: List[int],
    overlap: int,
    expected: Union[Type[Exception], List[np.ndarray]],
) -> None:
    perm = np.array(list(permutation))
    if isinstance(expected, list):
        output = _core.split(perm, dimensions, overlap)
        np.testing.assert_equal(output, expected)
    else:
        with pytest.raises(expected if not isinstance(expected, list) else None):
            _core.split(perm, dimensions, overlap)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# Based on https://github.com/aspuru-guzik-group/olympus

import numpy as np
from functools import partial
from nevergrad.parametrization import parameter as p
from ..base import ExperimentFunction

import nevergrad as ng


class OlympusSurface(ExperimentFunction):

    SURFACE_KINDS = (
        "Michalewicz",
        "AckleyPath",
        "Dejong",
        "HyperEllipsoid",
        "Levy",
        "Michalewicz",
        "Rastrigin",
        "Rosenbrock",
        "Schwefel",
        "StyblinskiTang",
        "Zakharov",
        "DiscreteAckley",
        "DiscreteDoubleWell",
        "DiscreteMichalewicz",
        "LinearFunnel",
        "NarrowFunnel",
        "GaussianMixture",
    )

    def __init__(
        self, kind: str, dimension: int = 10, noise_kind: str = "GaussianNoise", noise_scale: float = 1
    ) -> None:
        self.kind = kind
        self.param_dim = dimension
        self.noise_kind = noise_kind
        assert self.kind in OlympusSurface.SURFACE_KINDS
        assert self.noise_kind in ["GaussianNoise", "UniformNoise", "GammaNoise"]
        self.noise_scale = noise_scale
        self.surface = partial(self._simulate_surface, noise=True)
        self.surface_without_noise = partial(self._simulate_surface, noise=False)
        parametrization = p.Array(shape=(dimension,))
        parametrization.function.deterministic = False
        super().__init__(self.surface, parametrization)
        self.shift = self.parametrization.random_state.normal(size=self.dimension)

    def _simulate_surface(self, x: np.ndarray, noise: bool = True) -> float:
        try:
            from olympus.surfaces import import_surface  # pylint: disable=import-outside-toplevel
            from olympus import noises
        except ImportError as e:
            raise ng.errors.UnsupportedExperiment("Please install olympus for Olympus experiments") from e

        if noise:
            noise = noises.Noise(kind=self.noise_kind, scale=self.noise_scale)
            surface = import_surface(self.kind)(param_dim=self.param_dim, noise=noise)
        else:
            surface = import_surface(self.kind)(param_dim=self.param_dim)
        return surface.run(x - self.shift)[0][0]

    def evaluation_function(self, *recommendations) -> float:
        """Averages multiple evaluations if necessary"""
        x = recommendations[0].value
        return self.surface_without_noise(x - self.shift)


class OlympusEmulator(ExperimentFunction):
    DATASETS = (
        "suzuki",
        "fullerenes",
        "colors_bob",
        "photo_wf3",
        "snar",
        "alkox",
        "benzylation",
        "photo_pce10",
        "hplc",
        "colors_n9",
    )

    def __init__(self, dataset_kind: str = "alkox", model_kind: str = "NeuralNet") -> None:

        self.dataset_kind = dataset_kind
        self.model_kind = model_kind
        assert self.dataset_kind in OlympusEmulator.DATASETS
        assert self.model_kind in ["BayesNeuralNet", "NeuralNet"]
        parametrization = self._get_parametrization()
        parametrization.function.deterministic = False
        parametrization.set_name("")
        super().__init__(self._simulate_emulator, parametrization)

    def _get_parametrization(self) -> p.Parameter:
        try:
            from olympus.datasets import Dataset  # pylint: disable=import-outside-toplevel
        except ImportError as e:
            raise ng.errors.UnsupportedExperiment("Please install olympus for Olympus experiments") from e

        dataset = Dataset(self.dataset_kind)
        dimension = dataset.shape[1] - 1
        bounds = list(zip(*dataset.param_space.param_bounds))
        return p.Array(shape=(dimension,), lower=bounds[0], upper=bounds[1])

    def _simulate_emulator(self, x: np.ndarray) -> float:
        try:
            from olympus import Emulator  # pylint: disable=import-outside-toplevel
        except ImportError as e:
            raise ng.errors.UnsupportedExperiment("Please install olympus for Olympus experiments") from e

        emulator = Emulator(dataset=self.dataset_kind, model=self.model_kind)
        return emulator.run(x)[0][0] * (-1 if emulator.get_goal() == "maximize" else 1)
